import logging
import os
import pickle
from datetime import datetime
from enum import Enum
from uuid import UUID

import diskcache
from satcfdi.models import Code
from satcfdi.accounting import SatCFDI
from satcfdi.accounting.models import EstadoComprobante
from satcfdi.create.cfd.catalogos import MetodoPago, TipoDeComprobante
from satcfdi.pacs import sat

from .utils import estado_to_estatus

LIQUIDATED = 0
NOTIFIED = 2
STATUS_SAT = 3
FOLIO = 5
SERIE = 6
SERIE_PAGO = 7
SOLICITUDES = 'solicitudes'

sat_manager = sat.SAT()

logger = logging.getLogger(__name__)


class LocalDB(diskcache.Cache):
    def __init__(self, base_path: str):
        super().__init__(directory=os.path.join(base_path, 'cache'))
        self.base_path = base_path

    def folio(self) -> int:
        return self.get(FOLIO, 1000)

    def folio_set(self, value: int):
        self[FOLIO] = value

    def serie(self) -> str:
        return self.get(SERIE, 'A')

    def serie_set(self, value: str):
        self[SERIE] = value

    def serie_pago(self) -> str:
        return self.get(SERIE_PAGO, 'P')

    def serie_pago_set(self, value: str):
        self[SERIE_PAGO] = value

    def liquidated(self, uuid: UUID):
        return self.get((LIQUIDATED, uuid))

    def liquidated_set(self, uuid: UUID, value: bool):
        self[(LIQUIDATED, uuid)] = value

    def notified(self, uuid: UUID):
        return self.get((NOTIFIED, uuid))

    def notified_set(self, uuid: UUID, value: bool):
        self[(NOTIFIED, uuid)] = value

    def status(self, uuid: UUID):
        return self.get((STATUS_SAT, uuid), {})

    def status_merge(self, uuid: str | UUID, estatus: str, es_cancelable: str = None,
                     estatus_cancelacion: str = None, fecha_cancelacion: str = None, fecha_ultima_consulta: str = None):
        if isinstance(uuid, str):
            uuid = UUID(uuid)

        current_value = self.status(uuid)

        value = {
            "Estatus": Code(code=estatus, description=EstadoComprobante(estatus).name),
        }
        if es_cancelable:
            value["EsCancelable"] = es_cancelable

        if estatus_cancelacion:
            value["EstatusCancelacion"] = estatus_cancelacion

        if fecha_cancelacion:
            value["FechaCancelacion"] = datetime.fromisoformat(fecha_cancelacion)

        if fecha_ultima_consulta:
            fecha_ultima_consulta = datetime.fromisoformat(fecha_ultima_consulta)
            if current_value and current_value["UltimaConsulta"] > fecha_ultima_consulta:
                return
            value["UltimaConsulta"] = fecha_ultima_consulta
        else:
            value["UltimaConsulta"] = datetime.now().replace(microsecond=0)

        self[(STATUS_SAT, uuid)] = current_value | value

    def status_export(self, uuid: UUID):
        if i := self.status(uuid):
            return [
                uuid,
                i["Estatus"].code,
                i.get("EsCancelable"),
                i.get("EstatusCancelacion"),
                i.get("FechaCancelacion"),
                i.get("UltimaConsulta"),
            ]

    def get_solicitudes(self):
        return self.load_data(SOLICITUDES, {})

    def set_solicitudes(self, solicitudes):
        self.save_data(SOLICITUDES, solicitudes)

    def solicitud_merge(self, solicitud_id, response, request=None):
        solicitudes = self.get_solicitudes()
        solicitud = solicitudes.setdefault(solicitud_id, {})

        solicitud['response'] = solicitud.get('response', {}) | response
        if request:
            solicitud['request'] = solicitud.get('request', {}) | request
        solicitud['last_update'] = datetime.now().replace(microsecond=0)

        self.set_solicitudes(solicitudes)
        return solicitud

    def save_data(self, file, data):
        with open(os.path.join(self.base_path, file), 'wb') as f:
            pickle.dump(data, f)

    def load_data(self, file, default=None):
        try:
            with open(os.path.join(self.base_path, file), 'rb') as f:
                return pickle.load(f)
        except FileNotFoundError:
            return default


class StatusState(Enum):
    NONE = 1
    PAID = 2
    PENDING = 3
    IGNORED = 4
    CANCELLED = 5

    def __str__(self):
        if self.name == "NONE":
            return "💰"
        if self.name == "IGNORED":
            return "🚫"
        if self.name == "PAID":
            return "✔"
        if self.name == "PENDING":
            return ""
        if self.name == "CANCELLED":
            return "❌"


class LocalDBSatCFDI(LocalDB):
    def __init__(self, base_path, enviar_a_partir, pagar_a_partir):
        super().__init__(base_path)
        self.enviar_a_partir = enviar_a_partir
        self.pagar_a_partir = pagar_a_partir

    def notified(self, cfdi: SatCFDI):
        v = super().notified(cfdi.uuid)
        if v is None and cfdi["Fecha"] < self.enviar_a_partir:
            return True
        return v

    def notified_flip(self, cfdi: SatCFDI):
        v = not self.notified(cfdi)
        self.notified_set(cfdi.uuid, v)
        return v

    def liquidated(self, cfdi: SatCFDI):
        v = super().liquidated(cfdi.uuid)
        if v is None and cfdi["Fecha"] < self.pagar_a_partir[cfdi["MetodoPago"]]:
            return True
        return v

    def liquidated_flip(self, cfdi: SatCFDI):
        v = not self.liquidated(cfdi)
        self.liquidated_set(cfdi.uuid, v)
        return v

    def status_sat(self, cfdi: SatCFDI, update=False):
        if update:
            res = sat_manager.status(cfdi)
            if res["ValidacionEFOS"] == "200":
                self.status_merge(
                    uuid=cfdi.uuid,
                    estatus=estado_to_estatus(res["Estado"]),
                    es_cancelable=res["EsCancelable"],
                    estatus_cancelacion=res["EstatusCancelacion"]
                )
            else:
                raise ValueError("Error al actualizar estado de %s: %s", cfdi.uuid, res)
            return res
        else:
            return super().status(cfdi.uuid)

    def liquidated_state(self, cfdi: SatCFDI):
        if cfdi.estatus == EstadoComprobante.CANCELADO:
            return StatusState.CANCELLED

        if cfdi["TipoDeComprobante"] != TipoDeComprobante.INGRESO:
            return StatusState.NONE

        mpago = cfdi["MetodoPago"]
        if cfdi['Total'] == 0 or (mpago == MetodoPago.PAGO_EN_PARCIALIDADES_O_DIFERIDO and cfdi.saldo_pendiente == 0):
            return StatusState.PAID

        if self.liquidated(cfdi):
            if mpago == MetodoPago.PAGO_EN_PARCIALIDADES_O_DIFERIDO:
                return StatusState.IGNORED
            return StatusState.PAID

        return StatusState.PENDING



import logging
import re
import warnings

from bs4.builder import XMLParsedAsHTMLWarning
from satcfdi.models import RFC, RFCType
from satcfdi import csf
from satcfdi.pacs.sat import SAT

sat_service = SAT()
logger = logging.getLogger(__name__)
warnings.filterwarnings("ignore", category=XMLParsedAsHTMLWarning)

EMAIL_REGEX = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,7}\b'


def validar_client(client):
    errors = []

    rfc = client['Rfc']

    def error(msg):
        errors.append(f"{rfc}: {msg}")

    try:
        rfc = RFC(rfc)

        for email in client["Email"]:
            if not re.fullmatch(EMAIL_REGEX, email):
                error(f"Correo '{email}' is invalid")

        res = csf.retrieve(rfc, id_cif=client["IdCIF"])

        if rfc.type == RFCType.FISICA:
            razon_social = f"{res['Nombre']} {res['Apellido Paterno']} {res['Apellido Materno']}"
        elif 'Denominación o Razón Social' in res:
            razon_social = res['Denominación o Razón Social']
        else:
            error(f"Does not have 'Denominación o Razón Social'")

        if client['RazonSocial'] != razon_social:
            error(f"RazonSocial '{client['RazonSocial']}' is invalid, expected '{razon_social}'")

        if client['CodigoPostal'] != res['CP']:
            error(f"CodigoPostal '{client['CodigoPostal']}' is invalid, expected '{res['CP']}'")

        for r in res['Regimenes']:
            if r['RegimenFiscal'].code == None:
                error(f"RegimenFiscal '{r['RegimenFiscal']}' is invalid")

        if client['RegimenFiscal'] not in (r['RegimenFiscal'] for r in res['Regimenes']):
            regimen = ', '.join(r['RegimenFiscal'].code or "" for r in res['Regimenes'])
            error(
                f"RegimenFiscal '{client['RegimenFiscal']}' is invalid, "
                f"expected one of '{regimen}'"
            )

        if res['Situación del contribuyente'] not in ('ACTIVO', 'REACTIVADO'):
            error(f"Is not ACTIVO '{res['Situación del contribuyente']}'")

        taxpayer_status = sat_service.list_69b(rfc)
        if taxpayer_status:
            error(f"has status '{taxpayer_status}'")
    except Exception as ex:
        error(ex)

    return errors


def clientes_generar_txt(filename, clients):
    with open(filename, 'w', encoding='utf-8') as f:
        for i, (cliente_rfc, cliente_data) in enumerate(clients.items(), start=1):
            f.write(f"{i}|{cliente_rfc}|{cliente_data['RazonSocial']}|{cliente_data['CodigoPostal']}\n")



import yaml


class NoAliasDumper(yaml.SafeDumper):
    def ignore_aliases(self, data):
        return True


def print_yaml(data):
    print(
        to_yaml(data)
    )


def to_yaml(data):
    return yaml.dump(data, Dumper=NoAliasDumper, allow_unicode=True, width=1280, sort_keys=False)


def header_line(text):
    ln = (150 - len(text)) // 2
    return ("=" * ln) + " " + text + " " + ("=" * ln) + "\n"



import base64
import csv
import io
import itertools
import logging
import os
from datetime import date, datetime
from uuid import UUID
from zipfile import ZipFile

from PySimpleGUI import POPUP_BUTTONS_OK_CANCEL, PySimpleGUI as sg, POPUP_BUTTONS_NO_BUTTONS
from satcfdi import csf
from satcfdi.models import DatePeriod
from satcfdi.accounting import EmailManager
from satcfdi.accounting.models import EstadoComprobante
from satcfdi.accounting.process import complement_invoices
from satcfdi.create.cfd import cfdi40
from satcfdi.create.cfd.catalogos import MetodoPago, TipoDeComprobante
from satcfdi.exceptions import ResponseError
from satcfdi.pacs import Accept
from satcfdi.pacs.sat import SAT, EstadoSolicitud
from xlsxwriter.exceptions import XlsxFileError

from . import __version__, TEMP_DIRECTORY, ARCHIVOS_DIRECTORY, DATA_DIRECTORY, METADATA_FILE
from .client_validation import validar_client, clientes_generar_txt
from .environments import facturacion_environment
from .file_data_managers import ClientsManager, FacturasManager
from .gui_functions import generate_ingresos, pago_factura, exportar_facturas, archivos_folder, period_desc, parse_fecha_pago, parse_importe_pago, preview_cfdis, center_location, \
    CALENDAR_FECHA_FMT, ConsoleErrors, \
    generate_ajustes, generar_depositos, generate_pdf_template, calculate_declaracion_provisional
from .layout import make_layout, ActionButtonManager, TipoRecuperar, SearchOptions
from .localdb import LocalDBSatCFDI, StatusState
from .log_tools import header_line, print_yaml
from .mycfdi import MyCFDI
from .utils import random_string, to_date_period, load_certificate, to_int, cert_info, add_month, to_uuid, open_file, OS

logging.getLogger("weasyprint").setLevel(logging.ERROR)
logging.getLogger("fontTools").setLevel(logging.ERROR)

logger = logging.getLogger(__name__)


def open_launch_window():
    layout = [[sg.Text("New Window", key="new")]]
    window = sg.Window("Launch Window", layout, modal=True, size=(300, 300))
    window.read(timeout=1000)

    return window


class FacturacionGUI:
    def __init__(self, config):
        os.makedirs(TEMP_DIRECTORY, exist_ok=True)
        self.email_manager = EmailManager(
            **config['email']
        )
        pac = config['pac']
        pac_module, pac_class = pac['type'].split(".")
        mod = __import__(f"satcfdi.pacs.{pac_module}", fromlist=[pac_class])
        self.pac_service = getattr(mod, pac_class)(
            **pac['args']
        )
        self.csd_signer = load_certificate(config.get('csd')) if 'csd' in config else None
        self.fiel_signer = load_certificate(config.get('fiel')) if 'fiel' in config else None

        self.sat_service = SAT(signer=self.fiel_signer)
        self.rfc_prediales = config['rfc_prediales']

        self._all_invoices = None
        self.local_db = LocalDBSatCFDI(
            base_path=DATA_DIRECTORY,
            enviar_a_partir=config['enviar_a_partir'],
            pagar_a_partir=config['pagar_a_partir']
        )

        MyCFDI.local_db = self.local_db
        MyCFDI.base_dir = ARCHIVOS_DIRECTORY

        self.window = sg.Window(
            f"Facturación Mensual CFDI 4.0 {self.csd_signer.rfc}",
            make_layout(bool(self.fiel_signer), self.local_db),
            size=(1280, 720),
            resizable=True,
            font=("Courier New", 10, "bold"),
            ttk_theme="default",
            margins=(0, 0),
            # use_custom_titlebar=True,
            titlebar_font=("Courier New", 11, "bold"),
            finalize=True,
            scaling=config.get('scaling', 1.0),
        )
        self.has_focus = True

        self.action_button_manager = ActionButtonManager(
            button=self.window["crear_facturas"],
            preview=self.window["ver_preview"],
        )
        self.console = self.window["console"]
        self.set_inputs()

        self.window.bind("<FocusIn>", "_focus_in")
        self.window.bind("<FocusOut>", "_focus_out")

        for t in ('facturas_periodo', 'emitidas_search', 'recibidas_search', 'ajustes_periodo', 'depositos_periodo', 'serie', 'folio', 'serie_pago', 'periodo'):
            self.window[t].bind("<Return>", "_enter")
            self.window[t].bind("<FocusOut>", "_enter", propagate=False)

        modifier_key = "Command" if OS.get_os() == OS.MACOS else "Control"

        for t in ('facturas_table', 'clientes_table', 'emitidas_table', 'recibidas_table', 'correos_table', 'ajustes_table', 'depositos_table', 'solicitudes_table'):
            self.window[t].bind(f'<{modifier_key}-a>', '+select_all')
            self.window[t].bind('<BackSpace>', '+delete')  # BackSpace
            # self.window[t].bind('<Double-Button-1>', '_enter')
            self.window[t].bind('<Return>', '+enter')

    def run(self):
        self.main_loop()
        self.window.close()

    def initial_screen(self, emisor_cif):
        self.header("ACERCA DE")
        print_yaml({
            "version": __version__.__version__,
            "facturacion": "CFDI 4.0",
            "emisor": emisor_cif,
            "pac_service": {
                "Type": type(self.pac_service).__name__,
                "Rfc": self.pac_service.RFC,
                "Environment": str(self.pac_service.environment)
            },
            "fiel": cert_info(self.fiel_signer),
            "csd": cert_info(self.csd_signer),
        })

    def get_all_invoices(self):
        if not self._all_invoices:
            self._all_invoices = MyCFDI.get_all_cfdi()
        return self._all_invoices

    def add_created_invoice(self, invoice: MyCFDI):
        self._all_invoices[invoice.uuid] = invoice
        complement_invoices(self._all_invoices, invoice)

    def generate_invoice(self, invoice):
        ref_id = random_string()
        title = 'Generando factura'

        # Add Serie and Folio and signature
        folio = None
        if 'Serie' not in invoice:
            invoice['Serie'] = self.local_db.serie()
            folio = self.local_db.folio()
            invoice['Folio'] = str(folio)

        cfdi40.Comprobante.sign(invoice, self.csd_signer)

        attempts = 3
        try:
            for i in self.progress_iterate(
                    title, range(attempts), lambda r: f'Intentando de nuevo... Intento {r + 1} de {attempts}', skip_first=True, delay=2000
            ):
                try:
                    res = self.pac_service.stamp(
                        cfdi=invoice,
                        accept=Accept.XML_PDF,
                        ref_id=ref_id
                    )
                except Exception as ex:
                    self.show_console()
                    message = f"Error al generar factura: {invoice.get('Serie')}{invoice.get('Folio')} {invoice['Receptor']['Rfc']}"
                    message += f"\nIntento {i + 1} de {attempts}"

                    logger.exception(message)
                    print(message)
                    if isinstance(ex, ResponseError):
                        logger.error(f"Status Code: {ex.response.status_code}")
                        print(f"Status Code: {ex.response.status_code}")
                        logger.error(f"Response: {ex.response.text}")
                        print(f"Response: {ex.response.text}")
                    continue

                if folio is not None:
                    self.set_folio(folio + 1)
                cfdi = MyCFDI.move_to_folder(res.xml, pdf_data=res.pdf)
                self.add_created_invoice(cfdi)
                return cfdi
        finally:
            self.progress_cancel(title)

    def set_serie(self, serie: str = None):
        if serie:
            self.local_db.serie_set(serie.strip())
        self.set_inputs()

    def set_serie_pago(self, serie: str = None):
        self.local_db.serie_pago_set(serie.strip())
        self.set_inputs()

    def set_folio(self, folio: int = None):
        if folio:
            self.local_db.folio_set(folio)
        self.set_inputs()

    def set_inputs(self):
        folio = self.local_db.folio()
        serie = self.local_db.serie()
        serie_pago = self.local_db.serie_pago()

        self.window['serie'].update(serie)
        self.window['folio'].update(folio)
        self.window['serie_folio'].update(f"{serie}{folio}")

        self.window['serie_pago'].update(serie_pago)
        self.window['folio_pago'].update(
            "Se usara el mismo folio de la factura que se paga" if serie_pago else "Se usar serie y folio de proxima factura"
        )

    def nueva_solicitud(self, values):
        tipo_recuperar = values["tipo_recuperar"]

        args = {
            'fecha_inicial': datetime.strptime(values["fecha_inicial"], CALENDAR_FECHA_FMT),
            'fecha_final': datetime.strptime(values["fecha_final"], CALENDAR_FECHA_FMT),
            'rfc_receptor': self.sat_service.signer.rfc if tipo_recuperar == TipoRecuperar.Recibidas else None,
            'rfc_emisor': self.sat_service.signer.rfc if tipo_recuperar == TipoRecuperar.Emitidas else None,
            'tipo_solicitud': values["tipo_solicitud"],
        }

        response = self.sat_service.recover_comprobante_request(
            **args
        )

        self.local_db.solicitud_merge(response["IdSolicitud"], request=args, response=response)

    def recupera_comprobantes(self, response):
        if response["EstadoSolicitud"] == EstadoSolicitud.TERMINADA:
            for id_paquete in response['IdsPaquetes']:
                r, paquete = self.sat_service.recover_comprobante_download(
                    id_paquete=id_paquete
                )
                print(f"paquete: {id_paquete}")
                print_yaml(r)
                if paquete:
                    data = base64.b64decode(paquete)
                    with io.BytesIO(data) as b:
                        self.unzip_cfdi(b)

    def unzip_cfdi(self, file):
        title = 'Descomprimiendo'
        with ZipFile(file, "r") as zf:
            try:
                for fileinfo in self.progress_iterate(title, zf.infolist()):
                    data = zf.read(fileinfo)
                    match os.path.splitext(fileinfo.filename)[1]:
                        case ".xml":
                            self._all_invoices = None
                            MyCFDI.move_to_folder(data, pdf_data=None)
                        case ".pdf":
                            pass
                        case ".txt":
                            cfdi_metadata_reader = csv.reader(
                                (c.decode('utf-8') for c in data.splitlines() if c),
                                delimiter='~',
                                quotechar='|'
                            )
                            header = next(cfdi_metadata_reader)
                            for row in cfdi_metadata_reader:
                                row = dict(zip(header, row))
                                print_yaml(row)
                                self.local_db.status_merge(
                                    uuid=row['Uuid'],
                                    estatus=row['Estatus'],
                                    fecha_cancelacion=row['FechaCancelacion']
                                )
            finally:
                self.progress_cancel(title)

    def _read(self, timeout=0):
        event, values = self.window.read(timeout=timeout)
        if event in ("Exit", sg.WIN_CLOSED):
            return False
        return True

    def progress_iterate(self, title, items, fn=None, skip_first=False, delay=0):
        ln = len(items)
        try:
            for i, item in enumerate(items):
                if skip_first and i == 0:
                    pass
                else:
                    if not sg.one_line_progress_meter(
                            '',
                            i,
                            ln,
                            title,
                            fn(item) if callable(fn) else "",
                            key=title,
                            keep_on_top=True,
                            no_titlebar=True,
                            grab_anywhere=True,
                    ):
                        return

                if not self._read(timeout=delay * i):
                    return

                yield item
        finally:
            self.progress_cancel(title)

    @staticmethod
    def progress_cancel(title):
        sg.one_line_progress_meter_cancel(
            key=title
        )

    def action_button(self, action_name, action_items, action_text):
        try:
            match action_name:
                case 'solicitudes':
                    for solicitud in self.progress_iterate(action_text, action_items):
                        id_solicitud = solicitud["response"]["IdSolicitud"]
                        response = self.sat_service.recover_comprobante_status(
                            id_solicitud=id_solicitud
                        )
                        print_yaml(response)
                        self.local_db.solicitud_merge(id_solicitud, response=response)
                        self.recupera_comprobantes(response)

                case 'facturas' | 'pago':
                    for invoice in self.progress_iterate(action_text, action_items):
                        if not self.generate_invoice(invoice=invoice):
                            break

                case 'correos':
                    clients = ClientsManager()
                    emisor = clients[self.csd_signer.rfc]
                    with self.email_manager.sender as s:
                        for receptor, facturas, facturas_facturas_pendientes_meses_anteriores in self.progress_iterate(action_text, action_items):
                            def attachments():
                                for ni in facturas:
                                    yield ni.filename + ".xml"
                                    yield ni.filename + ".pdf"

                            s.send_email(
                                subject=f"Comprobantes Fiscales {receptor['RazonSocial']} - {receptor['Rfc']}",
                                to_addrs=receptor["Email"],
                                html=facturacion_environment.get_template('mail_facturas_template.html').render(
                                    facturas=facturas,
                                    facturas_pendientes_meses_anteriores=facturas_facturas_pendientes_meses_anteriores,
                                    emisor=emisor,
                                    receptor=receptor,
                                ),
                                file_attachments=attachments()
                            )
                            for r in facturas:
                                self.local_db.notified_set(r.uuid, True)

                case 'ajustes' | 'depositos':
                    with self.email_manager.sender as s:
                        grouped_action_items = []
                        for _, g_data in itertools.groupby(
                            sorted(
                                action_items,
                                key=lambda r: r["receptor"]["Rfc"]
                            ),
                            lambda r: r["receptor"]["Rfc"]
                        ):
                            grouped_action_items.append(list(g_data))

                        for g_data in self.progress_iterate(action_text, grouped_action_items):
                            file_names = []
                            for data in g_data:
                                if file := data['create_fn']():
                                    file_names.append(file)

                            if file_names:
                                receptor = g_data[0]['receptor']
                                if action_name == 'ajustes':
                                    subject = f"Ajuste Renta {receptor['RazonSocial']} - {receptor['Rfc']}"
                                elif action_name == 'depositos':
                                    subject = f"Depósito Renta {receptor['RazonSocial']} - {receptor['Rfc']}"
                                else:
                                    raise NotImplementedError()

                                s.send_email(
                                    subject=subject,
                                    to_addrs=receptor["Email"],
                                    html=facturacion_environment.get_template(f'mail_{action_name}_template.html').render(data),
                                    file_attachments=file_names
                                )

                case 'clientes':
                    for client in self.progress_iterate(
                            action_text, action_items, lambda x: f"Validando: {x['Rfc']}"
                    ):
                        if errors := validar_client(client):
                            self.show_console()
                            for e in errors:
                                print(e)

                case _:
                    raise ValueError(f"Invalid action: {action_name}")
        finally:
            self.progress_cancel(action_text)

    def set_selected_satcfdis_recibidas(self, cfdis: list):
        i = cfdis[0] if len(cfdis) == 1 else None

        if i:
            estatus = EstadoComprobante(i.estatus)
            self.window["status_sat_recibidas"].update(
                estatus.name.center(10),
                disabled=False,
                button_color="red4" if estatus != EstadoComprobante.VIGENTE else "dark green",
            )
        else:
            self.window["status_sat_recibidas"].update(
                "".ljust(10), disabled=True, button_color=sg.theme_background_color()
            )

    def set_selected_satcfdis(self, cfdis: list):
        i = cfdis[0] if len(cfdis) == 1 else None

        if i:
            estatus = EstadoComprobante(i.estatus)
            self.window["status_sat"].update(
                estatus.name.center(10),
                disabled=False,
                button_color="red4" if estatus != EstadoComprobante.VIGENTE else "dark green",
            )
        else:
            self.window["status_sat"].update(
                "".ljust(10), disabled=True, button_color=sg.theme_background_color()
            )

        # Email
        is_active = \
            bool(i) \
            and i["Emisor"]["Rfc"] == self.csd_signer.rfc \
            and i.estatus == EstadoComprobante.VIGENTE
        if is_active:
            self.window["email_notificada"].update(
                "Enviada".center(10) if self.local_db.notified(i) else "Por Enviar",
                disabled=False,
                button_color="dark green" if self.local_db.notified(i) else "red4",
            )
        else:
            self.window["email_notificada"].update(
                "".ljust(10), disabled=True, button_color=sg.theme_background_color()
            )

        # Pendiente de Pago
        is_pendientable = \
            is_active \
            and i["TipoDeComprobante"] == TipoDeComprobante.INGRESO \
            and (i["MetodoPago"] == MetodoPago.PAGO_EN_UNA_SOLA_EXHIBICION or i.saldo_pendiente) \
            and i["Total"]
        if is_pendientable:
            self.window["pendiente_pago"].update(
                (("Pagada" if i["MetodoPago"] == MetodoPago.PAGO_EN_UNA_SOLA_EXHIBICION else "Ignorada")
                 if self.local_db.liquidated(i) else "Por Pagar").center(10),
                disabled=False,
                button_color=("dark green" if i["MetodoPago"] == MetodoPago.PAGO_EN_UNA_SOLA_EXHIBICION else "yellow4")
                if self.local_db.liquidated(i) else "red4",
            )
        else:
            is_ppd_pagada = is_active \
                            and i["TipoDeComprobante"] == TipoDeComprobante.INGRESO \
                            and i["MetodoPago"] == MetodoPago.PAGO_EN_PARCIALIDADES_O_DIFERIDO \
                            and i.saldo_pendiente == 0
            if is_ppd_pagada:
                self.window["pendiente_pago"].update(
                    "Pagada".center(10),
                    disabled=True,
                    button_color="dark green",
                )
            else:
                self.window["pendiente_pago"].update(
                    "".ljust(10),
                    disabled=True,
                    button_color=sg.theme_background_color()
                )

        # PPD
        is_ppd_active = \
            is_active \
            and i["TipoDeComprobante"] == TipoDeComprobante.INGRESO \
            and i["MetodoPago"] == MetodoPago.PAGO_EN_PARCIALIDADES_O_DIFERIDO \
            and i.saldo_pendiente > 0

        self.window["ppd_action_items"].update(visible=is_ppd_active)
        self.window["importe_pago"].update(i.saldo_pendiente if is_ppd_active else '')
        if is_ppd_active:
            self.action_button_manager.set_items("pago", [i])
        else:
            self.action_button_manager.clear()

    def header(self, name, select_console=True):
        if select_console:
            self.show_console()
        self.console.update(header_line(name))
        self._read()

    def show_console(self):
        self.window['errores_tab'].select()

    def download_invoice(self, uuid: UUID):
        res = self.pac_service.recover(uuid, accept=Accept.XML_PDF)
        self._all_invoices = None
        return MyCFDI.move_to_folder(res.xml, pdf_data=res.pdf)

    def facturas_search(self):
        search_text = self.window["emitidas_search"].get()
        search_text = search_text.strip()

        if len(search_text) < 3:
            raise ValueError("Búsqueda debe de tener al menos 3 caracteres")

        def fact_iter():
            if search_text == SearchOptions.PorPagar:
                for i in self.get_all_invoices().values():
                    if i["Emisor"]["Rfc"] == self.csd_signer.rfc \
                            and self.local_db.liquidated_state(i) == StatusState.PENDING:
                        yield i
            elif search_text == SearchOptions.PorEnviar:
                for i in self.get_all_invoices().values():
                    if i["Emisor"]["Rfc"] == self.csd_signer.rfc \
                            and not self.local_db.notified(i) \
                            and i.estatus == EstadoComprobante.VIGENTE:
                        yield i
            elif date_search_text := to_date_period(search_text):
                for i in self.get_all_invoices().values():
                    if i["Emisor"]["Rfc"] == self.csd_signer.rfc \
                            and i["Fecha"] == date_search_text:
                        yield i
            elif uuid_search_text := to_uuid(search_text):
                if uuid_search_text not in self.get_all_invoices():
                    try:
                        self.download_invoice(uuid_search_text)
                    except ResponseError as e:
                        if e.response.status_code == 404:
                            self.error_message(f"Factura no encontrada en el PAC")
                        else:
                            raise e
                if c := self.get_all_invoices().get(uuid_search_text):
                    yield c
            else:
                up_search_text = search_text.upper()
                for i in self.get_all_invoices().values():
                    if i["Emisor"]["Rfc"] == self.csd_signer.rfc \
                            and (
                            i.name == up_search_text
                            or i["Receptor"]["Rfc"] == up_search_text
                            or up_search_text in i["Receptor"].get("Nombre", "")
                    ):
                        yield i

        self.window['emitidas_table'].update(
            values=sorted(fact_iter(), key=lambda x: x["Fecha"], reverse=False),
        )

    def facturas_search_recibidas(self):
        search_text = self.window["recibidas_search"].get()
        search_text = search_text.strip()

        if len(search_text) < 3:
            raise ValueError("Búsqueda debe de tener al menos 3 caracteres")

        def fact_iter():
            if search_text == SearchOptions.PorPagar:
                for i in self.get_all_invoices().values():
                    if i["Receptor"]["Rfc"] == self.csd_signer.rfc \
                            and self.local_db.liquidated_state(i) == StatusState.PENDING:
                        yield i
            elif search_text == SearchOptions.PorEnviar:
                for i in self.get_all_invoices().values():
                    if i["Receptor"]["Rfc"] == self.csd_signer.rfc \
                            and not self.local_db.notified(i) \
                            and i.estatus == EstadoComprobante.VIGENTE:
                        yield i
            elif date_search_text := to_date_period(search_text):
                for i in self.get_all_invoices().values():
                    if i["Receptor"]["Rfc"] == self.csd_signer.rfc \
                            and i["Fecha"] == date_search_text:
                        yield i
            elif uuid_search_text := to_uuid(search_text):
                if uuid_search_text not in self.get_all_invoices():
                    try:
                        self.download_invoice(uuid_search_text)
                    except ResponseError as e:
                        if e.response.status_code == 404:
                            self.error_message(f"Factura no encontrada en el PAC")
                        else:
                            raise e
                if c := self.get_all_invoices().get(uuid_search_text):
                    yield c
            else:
                up_search_text = search_text.upper()
                for i in self.get_all_invoices().values():
                    if i["Receptor"]["Rfc"] == self.csd_signer.rfc \
                            and (
                            i.name == up_search_text
                            or i["Emisor"]["Rfc"] == up_search_text
                            or up_search_text in i["Emisor"].get("Nombre", "")
                    ):
                        yield i

        self.window['recibidas_table'].update(
            values=sorted(fact_iter(), key=lambda x: x["Fecha"], reverse=False),
        )

    def crear_pago(self, values, facturas_pagar):
        clients = ClientsManager()
        fecha_pago = parse_fecha_pago(values["fecha_pago"])

        max_dias = 30
        dif_dias = (datetime.now() - fecha_pago).days
        if dif_dias > max_dias:
            res = sg.popup(
                f"Estas seguro que quieres hacer Complemento de Pago con fecha de pago de hace '{dif_dias}' dias?",
                title="Confirmar",
                button_type=POPUP_BUTTONS_OK_CANCEL,
                location=center_location(self.window),
                keep_on_top=True,
            )
            if res != "OK":
                return []

        importe_pago = parse_importe_pago(values["importe_pago"])
        self.window["importe_pago"].update(importe_pago)

        # noinspection PyUnresolvedReferences
        cfdi = pago_factura(
            receptor_cif=clients[facturas_pagar[0]["Receptor"]["Rfc"]],
            factura_pagar=facturas_pagar[0],
            fecha_pago=fecha_pago,
            forma_pago=values["forma_pago"],
            importe_pago=importe_pago,
            serie_pago=self.local_db.serie_pago(),
        )
        return [cfdi]

    def error_message(self, ex):
        sg.Popup(
            ex,
            no_titlebar=True,
            background_color="red4",
            location=center_location(self.window),
            button_type=POPUP_BUTTONS_NO_BUTTONS,
            auto_close=True,
            non_blocking=True,
        )

    def done_message(self, ex):
        sg.Popup(
            ex,
            no_titlebar=True,
            background_color="green4",
            location=center_location(self.window),
            button_type=POPUP_BUTTONS_NO_BUTTONS,
            auto_close=True,
            non_blocking=True,
        )

    def nuevas_facturas(self, values, force=False):
        facturas_table = self.window['facturas_table']
        has_value = bool(facturas_table.metadata)
        facturas_table.update(values=[])
        self.window['preparar_facturas_text'].update("")

        if has_value or force:
            dp = to_date_period(values["facturas_periodo"])
            if dp is None or dp.month is None:
                raise ValueError("Periodo no válido")

            self.window['preparar_facturas_text'].update(f"{period_desc(dp)}")

            cfdis = generate_ingresos(
                clients=ClientsManager(),
                facturas=FacturasManager(dp)["Facturas"],
                dp=dp,
                emisor_rfc=self.csd_signer.rfc
            )
            facturas_table.update(
                values=cfdis,
            )

    def nuevos_ajustes(self, values, force=False):
        ajustes_table = self.window['ajustes_table']
        has_value = bool(ajustes_table.metadata)
        ajustes_table.update(values=[])
        self.window['preparar_ajustes_text'].update("")

        if has_value or force:
            dp = to_date_period(values["ajustes_periodo"])
            if dp is None or dp.month is None:
                raise ValueError("Periodo no válido")

            dp_effective = add_month(dp, 1)
            self.window['preparar_ajustes_text'].update(f"{period_desc(dp)}, Ajustes Efectivos Al: {period_desc(dp_effective)}")

            ajustes = generate_ajustes(
                clients=ClientsManager(),
                facturas=FacturasManager(None)["Facturas"],
                dp=dp,
                dp_effective=dp_effective,
                emisor_rfc=self.csd_signer.rfc
            )
            ajustes_table.update(
                values=ajustes,
            )

    def nuevos_depositos(self, values, force=False):
        depositos_table = self.window['depositos_table']
        has_value = bool(depositos_table.metadata)
        depositos_table.update(values=[])
        self.window['preparar_depositos_text'].update("")

        if has_value or force:
            dp = to_date_period(values["depositos_periodo"])
            if dp is None or dp.month is None:
                raise ValueError("Periodo no válido")

            self.window['preparar_depositos_text'].update(f"{period_desc(dp)}")

            depositos = generar_depositos(
                clients=ClientsManager(),
                facturas=FacturasManager(dp)["Facturas"],
                dp=dp,
                emisor_rfc=self.csd_signer.rfc,

            )
            depositos_table.update(
                values=depositos,
            )

    def main_tab_group(self, values):
        self.action_button_manager.clear()

        match values['main_tab_group']:
            case 'facturas_tab':
                self.nuevas_facturas(values)

            case 'clientes_tab':
                self.window['clientes_table'].update(
                    values=list(ClientsManager().values()),
                )

            case 'emitidas_tab':
                self.facturas_search()

            case 'recibidas_tab':
                self.facturas_search_recibidas()

            case 'correos_tab':
                now = date.today()
                dp_now = DatePeriod(now.year, now.month)
                clients = ClientsManager()

                def correos():
                    for receptor_rfc, notify_invoices in itertools.groupby(
                            sorted(
                                (i for i in self.get_all_invoices().values()
                                 if i["Emisor"]["Rfc"] == self.csd_signer.rfc
                                    and i.estatus == EstadoComprobante.VIGENTE
                                    and not self.local_db.notified(i)
                                 ),
                                key=lambda r: r["Receptor"]["Rfc"]
                            ),
                            lambda r: r["Receptor"]["Rfc"]
                    ):
                        notify_invoices = list(notify_invoices)

                        def fac_pen_iter():
                            for i in self.get_all_invoices().values():
                                if i["Emisor"]["Rfc"] == self.csd_signer.rfc \
                                        and self.local_db.liquidated_state(i) == StatusState.PENDING \
                                        and i["Fecha"] < dp_now \
                                        and i["Receptor"]["Rfc"] == receptor_rfc \
                                        and i not in notify_invoices:
                                    yield i

                        fac_pen = sorted(
                            fac_pen_iter(),
                            key=lambda r: r["Fecha"]
                        )
                        yield clients[receptor_rfc], notify_invoices, fac_pen

                self.window['correos_table'].update(
                    values=list(correos()),
                )

            case 'ajustes_tab':
                self.nuevos_ajustes(values)

            case 'depositos_tab':
                self.nuevos_depositos(values)

            case 'solicitudes_tab':
                solitudes = self.local_db.get_solicitudes()
                self.window['solicitudes_table'].update(
                    values=list(solitudes.values()),
                )

            case 'contabilidad_tab':
                pass


    def main_loop(self):
        _, values = self.window.read(timeout=0)
        event = "main_tab_group"

        while True:
            if event in ("Exit", sg.WIN_CLOSED):
                return

            self.action(event, values)
            event, values = self.window.read()

    def action(self, event, values):
        try:
            match event:
                case '_focus_in':
                    if not self.has_focus:
                        self.has_focus = True
                        if values["main_tab_group"] in ("clientes_tab", "facturas_tab", "ajustes_tab", "correos_tab"):
                            self.main_tab_group(values)

                case '_focus_out':
                    try:
                        self.has_focus = bool(self.window.TKroot.focus_get())
                    except KeyError:
                        self.has_focus = True

                case "folio_enter":
                    self.set_folio(to_int(values["folio"]))

                case "serie_enter":
                    self.set_serie(values["serie"])

                case "serie_pago_enter":
                    self.set_serie_pago(values["serie_pago"])

                case "about":
                    clients = ClientsManager()
                    self.initial_screen(clients[self.csd_signer.rfc])

                case "nueva_solicitud":
                    self.nueva_solicitud(values)
                    self.main_tab_group(values)

                case "buscar_facturas":
                    self.window["emitidas_search"].update(values["buscar_facturas"])
                    self.facturas_search()

                case "buscar_facturas_recibidas":
                    self.window["recibidas_search"].update(values["buscar_facturas_recibidas"])
                    self.facturas_search_recibidas()

                case "emitidas_search_enter":
                    self.facturas_search()

                case "recibidas_search_enter":
                    self.facturas_search_recibidas()

                case "facturas_periodo_enter":
                    self.nuevas_facturas(values, force=True)

                case "ajustes_periodo_enter":
                    self.nuevos_ajustes(values, force=True)

                case "depositos_periodo_enter":
                    self.nuevos_depositos(values, force=True)

                case 'main_tab_group':
                    self.main_tab_group(values)

                case 'facturas_table+enter':
                    # noinspection PyUnresolvedReferences
                    if s_items := self.window["facturas_table"].selected_items():
                        preview_cfdis(s_items)

                case 'clientes_table+enter':
                    # noinspection PyUnresolvedReferences
                    for client in self.window["clientes_table"].selected_items():
                        url = csf.url(rfc=client["Rfc"], id_cif=client["IdCIF"])
                        open_file(url)

                case 'emitidas_table+enter' | 'recibidas_table+enter':
                    # noinspection PyUnresolvedReferences
                    table = event.split("+")[0]
                    if s_items := self.window[table].selected_items():
                        preview_cfdis(s_items)

                case 'ajustes_table+enter' | 'depositos_table+enter':
                    table = event.split("+")[0]
                    # noinspection PyUnresolvedReferences
                    for items in self.window[table].selected_items():
                        if file_name := items['create_fn']():
                            open_file(
                                os.path.abspath(file_name)
                            )
                        else:
                            self.error_message("No se pudo crear el archivo")

                case 'correos_table+enter' | 'solicitudes_table+enter':
                    pass

                case "facturas_table" | "clientes_table" | "correos_table" | "ajustes_table" | "depositos_table" | "emitidas_table" | "recibidas_table" | "solicitudes_table" | "depositos_table":
                    # noinspection PyUnresolvedReferences
                    s_items = self.window[event].selected_items()
                    if event == "emitidas_table":
                        self.set_selected_satcfdis(s_items)
                    elif event == "recibidas_table":
                        self.set_selected_satcfdis_recibidas(s_items)
                    else:
                        self.action_button_manager.set_items(event.split("_")[0], s_items)

                case "facturas_table+select_all" | "clientes_table+select_all" | "correos_table+select_all" | \
                     "ajustes_table+select_all" | "emitidas_table+select_all" | "recibidas_table" | 'solicitudes_table+select_all' | \
                     'depositos_table+select_all':
                    # noinspection PyUnresolvedReferences
                    self.window[event.split("+")[0]].select_all()

                case "facturas_table+delete" | "clientes_table+delete" | "correos_table+delete" | \
                     "ajustes_table+delete" | "emitidas_table+delete" | "recibidas_table+delete":
                    # noinspection PyUnresolvedReferences
                    # self.window[event.split("+")[0]].delete_selected()
                    pass

                case "solicitudes_table+delete":
                    solitudes = self.local_db.get_solicitudes()
                    # noinspection PyUnresolvedReferences
                    sel = self.window["solicitudes_table"].selected_items()
                    for s in sel:
                        if s["response"].get("EstadoSolicitud").code < EstadoSolicitud.TERMINADA:
                            self.error_message("No se puede eliminar una solicitud en proceso")
                            continue
                        del solitudes[s["response"]["IdSolicitud"]]
                    self.local_db.set_solicitudes(solitudes)
                    self.main_tab_group(values)

                case "status_sat_recibidas":
                    # noinspection PyUnresolvedReferences
                    if i := self.window["recibidas_table"].selected_items()[0]:
                        res = self.local_db.status_sat(i, update=True)
                        self.done_message(f"Estado: {res['Estado']}")
                        self.set_selected_satcfdis_recibidas([i])
                        # noinspection PyUnresolvedReferences
                        self.window['recibidas_table'].refresh()

                case "status_sat":
                    # noinspection PyUnresolvedReferences
                    if i := self.window["emitidas_table"].selected_items()[0]:
                        res = self.local_db.status_sat(i, update=True)
                        self.done_message(f"Estado: {res['Estado']}")
                        self.set_selected_satcfdis([i])
                        # noinspection PyUnresolvedReferences
                        self.window['emitidas_table'].refresh()

                case "pendiente_pago":
                    # noinspection PyUnresolvedReferences
                    if i := self.window["emitidas_table"].selected_items()[0]:
                        self.local_db.liquidated_flip(i)
                        self.set_selected_satcfdis([i])
                        # noinspection PyUnresolvedReferences
                        self.window['emitidas_table'].refresh()

                case "email_notificada":
                    # noinspection PyUnresolvedReferences
                    if i := self.window["emitidas_table"].selected_items()[0]:
                        self.local_db.notified_flip(i)
                        self.set_selected_satcfdis([i])
                        # noinspection PyUnresolvedReferences
                        self.window['emitidas_table'].refresh()

                case "crear_facturas" | "ver_preview":
                    action_text = self.action_button_manager.text()
                    action_name = self.action_button_manager.name
                    action_items = self.action_button_manager.items

                    if action_name == "pago":
                        action_items = self.crear_pago(values, action_items)
                        if not action_items:
                            return

                    if event == "ver_preview":
                        if action_name in ("facturas", "pago"):
                            preview_cfdis(action_items)

                    elif event == "crear_facturas":
                        res = sg.popup(
                            f"Estas seguro que quieres '{action_text}'?",
                            title="Confirmar",
                            button_type=POPUP_BUTTONS_OK_CANCEL,
                            location=center_location(self.window),
                            keep_on_top=True,
                        )
                        if res == "OK":
                            self.header(action_text.upper(), select_console=False)
                            self.action_button_manager.clear()
                            self.action_button(
                                action_name=action_name,
                                action_items=action_items,
                                action_text=action_text
                            )
                            self.main_tab_group(values)

                case "editar_clientes":
                    open_file(
                        os.path.abspath("clientes.yaml")
                    )

                case "exportar_clientes":
                    filename = 'clientes.txt'
                    clients = ClientsManager()
                    clientes_generar_txt(filename, clients)
                    open_file(
                        os.path.dirname(
                            os.path.abspath(filename)
                        )
                    )

                case "editar_facturas" | "editar_ajustes":
                    open_file(
                        os.path.abspath("facturas.yaml")
                    )

                case "editar_configurar":
                    open_file(
                        os.path.abspath("config.yaml")
                    )

                case "ver_config":
                    open_file(
                        os.path.abspath(".")
                    )

                case "ver_excel":
                    dp = to_date_period(values["periodo"])
                    clients = ClientsManager()
                    emisor_cif = clients[self.csd_signer.rfc]
                    archivo_excel = exportar_facturas(
                        self.get_all_invoices(),
                        dp,
                        emisor_cif,
                        self.rfc_prediales
                    )
                    open_file(
                        os.path.abspath(archivo_excel)
                    )

                case "periodo_enter":
                    dp = to_date_period(values["periodo"])
                    clients = ClientsManager()
                    emisor_cif = clients[self.csd_signer.rfc]

                    declaracion_provisional = calculate_declaracion_provisional(
                        self.get_all_invoices(),
                        dp,
                        emisor_cif,
                        self.rfc_prediales
                    )
                    self.window['declaracion'].update(
                        append=False,
                        value=declaracion_provisional
                    )

                case "ver_carpeta":
                    dp = to_date_period(values["periodo"])
                    directory = archivos_folder(dp)
                    open_file(
                        os.path.abspath(directory)
                    )

                case "organizar_facturas":
                    MyCFDI.rename_invoices(search_path="**/*.xml")

                case "cargar_zip":
                    zip_file = sg.popup_get_file('', multiple_files=False, no_window=True, file_types=(("ZIP Files", "*.zip"),))
                    if zip_file:
                        self.unzip_cfdi(zip_file)

                case 'importar_emitidas':
                    csv_file = sg.popup_get_file('', multiple_files=False, no_window=True, file_types=(("CSV Files", "*.csv"),))
                    if csv_file:
                        all_invoices = self.get_all_invoices()
                        with open(csv_file, newline='', encoding='utf-8') as f:
                            reader = csv.reader(f)
                            header = next(reader)
                            for row in reader:
                                row = dict(zip(header, row))
                                uuid = UUID(row["Folio Fiscal (UUID)"])
                                if uuid not in all_invoices:
                                    cfdi = self.download_invoice(uuid)

                                    if row.get("Estatus", "Entregado SAT") != "Entregado SAT":
                                        self.local_db.status_sat(cfdi, update=True)
                        self.done_message("FIN")

                case "exportar_metadata":
                    with open(METADATA_FILE, 'w', newline='', encoding='utf-8') as f:
                        writer = csv.writer(f)
                        for i in self.get_all_invoices():
                            if status := self.local_db.status_export(i):
                                writer.writerow(status)
                    self.done_message("FIN")

                case "importar_metadata":
                    with open(METADATA_FILE, newline='', encoding='utf-8') as f:
                        reader = csv.reader(f)
                        for row in reader:
                            self.local_db.status_merge(*row)
                    self.done_message("FIN")

                case _:
                    logger.error(f"Unknown event '{event}'")

        except (ValueError, XlsxFileError) as ex:
            self.error_message(ex)
        except ConsoleErrors as ex:
            self.header(str(ex))
            for error in ex.errors:
                self.console.update(f"{error}\n", append=True)
        except Exception as ex:
            self.header("Exception")
            self.console.update(append=True, value=str(ex))
            logger.exception("Main Loop Exception")


from datetime import datetime

from satdigitalinvoice.localdb import LocalDB, LocalDBSatCFDI
from satdigitalinvoice.utils import random_string


def test_localdb():
    db = LocalDBSatCFDI(
        base_path=".data",
        enviar_a_partir=datetime(2020, 1, 1),
        pagar_a_partir={
            "PUE": datetime(2020, 1, 1),
            "PPD": datetime(2020, 1, 1),
        }
    )

    db.folio_set(10)
    assert db.folio() == 10

    a = random_string()

    db.save_data('test_save_data', a)
    assert db.load_data('test_save_data') == a





import logging
import os
from datetime import datetime, date
from decimal import Decimal
from decimal import InvalidOperation

import xlsxwriter
from markdown2 import markdown
from satcfdi import render
from satcfdi.accounting import filter_invoices_iter, filter_payments_iter, invoices_export, payments_export
from satcfdi.accounting.process import payments_groupby_receptor, payments_retentions_export
from satcfdi.create.cfd import cfdi40
from satcfdi.create.cfd.catalogos import Impuesto
from satcfdi.create.cfd.cfdi40 import Comprobante, PagoComprobante
from satcfdi.models import DatePeriod
from satcfdi.pacs import sat
try:
    from weasyprint import HTML, CSS
except:
    pass

from . import SOURCE_DIRECTORY, ARCHIVOS_DIRECTORY, TEMP_DIRECTORY
from .environments import facturacion_environment
from .exceptions import ConsoleErrors
from .formatting_functions.common import fecha_mes
from .log_tools import to_yaml
from .utils import add_month, find_best_match, months_between, open_file

logger = logging.getLogger(__name__)
logger.level = logging.INFO

sat_manager = sat.SAT()

PERIODICIDAD = {
    "Mensual": 1,
    "Bimestral": 2,
    "Trimestral": 3,
    "Cuatrimestral": 4,
    "Semestral": 6,
    "Anual": 12
}
CALENDAR_FECHA_FMT = '%Y-%m-%d'


def create_cfdi(receptor_cif, factura_details, emisor_cif):
    emisor = cfdi40.Emisor(
        rfc=emisor_cif['Rfc'],
        nombre=emisor_cif['RazonSocial'],
        regimen_fiscal=emisor_cif['RegimenFiscal']
    )
    emisor = emisor | factura_details.get('Emisor', {})
    invoice = cfdi40.Comprobante(
        emisor=emisor,
        lugar_expedicion=emisor_cif['CodigoPostal'],
        receptor=cfdi40.Receptor(
            rfc=receptor_cif['Rfc'],
            nombre=receptor_cif['RazonSocial'],
            uso_cfdi=factura_details['UsoCFDI'],
            domicilio_fiscal_receptor=receptor_cif['CodigoPostal'],
            regimen_fiscal_receptor=receptor_cif['RegimenFiscal']
        ),
        metodo_pago=factura_details['MetodoPago'],
        forma_pago=factura_details['FormaPago'],
        # cfdi_relacionados=[
        #     cfdi40.CfdiRelacionados(
        #         tipo_relacion=TipoRelacion.SUSTITUCION_DE_LOS_CFDI_PREVIOSaaaaaaa,
        #         cfdi_relacionado=''xyc',
        #     )
        # ],
        # serie=serie,
        # folio=folio,
        conceptos=factura_details["Conceptos"]
    )
    invoice = invoice.process()
    return invoice


def parse_periodo_mes_ajuste(periodo_mes_ajuste: str):
    parts = periodo_mes_ajuste.split(".")
    if len(parts) != 2:
        raise ValueError("Periodo Invalido")

    periodo, mes_ajuste = parts
    if not mes_ajuste.isnumeric():
        raise ValueError("Periodo Invalido")

    mes_ajuste = int(mes_ajuste)
    if not (12 >= int(mes_ajuste) >= 1):
        raise ValueError("Periodo Invalido")

    if periodo not in PERIODICIDAD:
        raise ValueError("Periodo Invalido")

    return PERIODICIDAD[periodo], mes_ajuste


def format_concepto_desc(concepto, periodo):
    concepto = concepto.copy()
    template = facturacion_environment.from_string(concepto["Descripcion"])
    concepto["Descripcion"] = template.render(
        periodo=periodo
    )
    return concepto


def period_desc(dp: DatePeriod):
    return fecha_mes(date(year=dp.year, month=dp.month, day=1))


def periodicidad_desc(dp: DatePeriod, periodo_mes_ajuste, offset):
    periodo_meses, mes_ajuste = parse_periodo_mes_ajuste(periodo_mes_ajuste)

    if (dp.month - mes_ajuste) % periodo_meses == 0:
        if offset:
            dp = add_month(dp, offset)

        periodo = period_desc(dp)
        if periodo_meses > 1:
            periodo += " AL " + period_desc(
                dp=add_month(dp, periodo_meses - 1)
            )

        return periodo
    return None


def generate_ingresos(clients, facturas, dp, emisor_rfc):
    emisor_cif = clients[emisor_rfc]
    errors = []

    def facturas_iter():
        for i, f in enumerate(facturas, start=1):
            try:
                def prepare_concepto(c):
                    periodo = periodicidad_desc(
                        dp,
                        c['_periodo_mes_ajuste'],
                        c.get('_desfase_mes')
                    )
                    if periodo and c['ValorUnitario'] is not None:
                        return format_concepto_desc(c, periodo=periodo)

                receptor_cif = clients.get(f['Receptor'])
                if not receptor_cif:
                    raise ValueError("client not found")

                if f["MetodoPago"] == "PPD" and f["FormaPago"] != "99":
                    raise ValueError(f"FormaPago '{f['FormaPago']}' is invalid, expected '99' for PPD")

                if conceptos := [x for x in (prepare_concepto(c) for c in f["Conceptos"]) if x]:
                    f["Conceptos"] = conceptos
                    cfdi = create_cfdi(receptor_cif, f, emisor_cif)

                    expected_total = f.get('Total')
                    if expected_total is not None and expected_total != cfdi['Total']:
                        raise ValueError(f"Total '{expected_total}' is invalid, expected '{cfdi['Total']}'")

                    yield cfdi
            except Exception as e:
                errors.append(f"{i} {f['Receptor']}: {str(e)}")

    cfdis = list(facturas_iter())
    if errors:
        raise ConsoleErrors("Generar Facturas Errores", errors=errors)
    return cfdis


def parse_fecha_pago(fecha_pago):
    if not fecha_pago:
        raise ValueError("Fecha de Pago es requerida")

    fecha_pago = datetime.strptime(fecha_pago, CALENDAR_FECHA_FMT)
    if fecha_pago > datetime.now():
        raise ValueError("Fecha de Pago es mayor a la fecha actual")

    if fecha_pago.replace(hour=12) > datetime.now():
        fecha_pago = datetime.now()
    else:
        fecha_pago = fecha_pago.replace(hour=12)

    return fecha_pago


def parse_importe_pago(importe_pago: str):
    try:
        return round(Decimal(importe_pago), 2)
    except InvalidOperation:
        raise ValueError("Importe de Pago es invalido")


def pago_factura(factura_pagar, fecha_pago: datetime, forma_pago: str, importe_pago: Decimal = None, serie_pago=None, receptor_cif=None):
    c = factura_pagar

    receptor = cfdi40.Receptor(
        rfc=receptor_cif['Rfc'],
        nombre=receptor_cif['RazonSocial'],
        domicilio_fiscal_receptor=receptor_cif['CodigoPostal'],
        regimen_fiscal_receptor=receptor_cif['RegimenFiscal'],
        uso_cfdi="CP01"
    )

    invoice = Comprobante.pago_comprobantes(
        receptor=receptor,
        serie=serie_pago or None,
        folio=c.get('Folio') if serie_pago else None,
        comprobantes=[
            PagoComprobante(
                comprobante=c,
                num_parcialidad=c.ultima_num_parcialidad + 1,
                imp_saldo_ant=c.saldo_pendiente,
                imp_pagado=importe_pago
            )
        ],
        fecha_pago=fecha_pago,
        forma_pago=forma_pago,
    )
    return invoice.process()


def find_ajustes(facturas, mes_ajuste):
    for f in facturas:
        rfc = f["Receptor"]
        for concepto in f["Conceptos"]:
            _, mes_aj = parse_periodo_mes_ajuste(concepto['_periodo_mes_ajuste'])
            if mes_aj == mes_ajuste:
                yield rfc, concepto


def generate_ajustes(clients, facturas, dp, dp_effective, emisor_rfc):
    errors = []
    ajustes_dir = os.path.join(archivos_folder(dp), 'ajustes')
    os.makedirs(ajustes_dir, exist_ok=True)

    def ajustes_iter():
        for i, (receptor_rfc, concepto) in enumerate(find_ajustes(facturas, dp_effective.month), start=1):
            try:
                valor_unitario_raw = concepto["ValorUnitario"]

                if isinstance(valor_unitario_raw, dict):
                    vu_eff, vu = find_best_match(valor_unitario_raw, dp)
                    vun_eff, vun = find_best_match(valor_unitario_raw, dp_effective)
                    if vu_eff == vun_eff:
                        vun = None
                        num_meses = None
                    else:
                        num_meses = months_between(vun_eff, vu_eff)

                    if vun and vu:
                        ajuste_porcentaje = round((vun / vu - 1) * 100, 2)
                    else:
                        ajuste_porcentaje = None
                else:
                    vu = valor_unitario_raw
                    vun = None
                    num_meses = None
                    ajuste_porcentaje = None

                concepto = format_concepto_desc(concepto, periodo="INMUEBLE")
                file_name = os.path.join(ajustes_dir, f'AjusteRenta_{receptor_rfc}_{i}.pdf')
                client_receptor = clients[receptor_rfc]  # type: dict
                data = {
                    "receptor": client_receptor,
                    "emisor": clients[emisor_rfc],
                    "concepto": concepto,
                    "valor_unitario": vu,
                    "valor_unitario_nuevo": vun or '',
                    "ajuste_porcentaje": ajuste_porcentaje or "",
                    "meses": num_meses or '',
                    "efectivo_periodo_desc": periodicidad_desc(dp_effective, concepto['_periodo_mes_ajuste'], concepto.get('_desfase_mes')),
                    "periodo": concepto['_periodo_mes_ajuste'].split('.')[0].upper(),
                }
                data['create_fn'] = create_ajuste_fn(ajuste_porcentaje, data, file_name)

                yield data
            except Exception as e:
                errors.append(f"{i} {receptor_rfc}: {str(e)}")

    cfdis = list(ajustes_iter())
    if errors:
        raise ConsoleErrors("Generar Ajustes Errores", errors=errors)
    return cfdis


def create_ajuste_fn(ajuste_porcentaje, data, file_name):
    def fn():
        if ajuste_porcentaje:
            res = generate_pdf_template(
                template_name='ajuste_template.md',
                fields=data
            )
            with open(file_name, 'wb') as f:
                f.write(res)
            return file_name

    return fn


def generar_depositos(clients, facturas, dp, emisor_rfc):
    errors = []
    depositos_dir = os.path.join(archivos_folder(dp), 'depositos')
    os.makedirs(depositos_dir, exist_ok=True)

    def depositos_iter():
        for i, (receptor_rfc, concepto) in enumerate(find_ajustes(facturas, dp.month), start=1):
            try:
                vu = concepto["ValorUnitario"]

                concepto = format_concepto_desc(concepto, periodo="INMUEBLE")
                file_name = os.path.join(depositos_dir, f'Deposito_{receptor_rfc}_{i}.pdf')

                client_receptor = clients[receptor_rfc]  # type: dict
                data = {
                    "receptor": client_receptor,
                    "emisor": clients[emisor_rfc],
                    "concepto": concepto,
                    "valor_unitario": vu,
                    "periodo": concepto['_periodo_mes_ajuste'].split('.')[0].upper(),
                }
                data['create_fn'] = create_deposito_fn(data, file_name)
                yield data
            except Exception as e:
                errors.append(f"{i} {receptor_rfc}: {str(e)}")

    cfdis = list(depositos_iter())
    if errors:
        raise ConsoleErrors("Generar Depositos Errores", errors=errors)
    return cfdis


def create_deposito_fn(data, file_name):
    def fn():
        res = generate_pdf_template(
            template_name='deposito_template.md',
            fields=data
        )
        with open(file_name, 'wb') as f:
            f.write(res)
        return file_name

    return fn


def archivos_folder(dp: DatePeriod):
    if dp.month:
        return os.path.join(ARCHIVOS_DIRECTORY, str(dp.year), str(dp.year) + "-{:02d}".format(dp.month))
    return os.path.join(ARCHIVOS_DIRECTORY, str(dp.year))


def archivos_filename(dp: DatePeriod, ext="xlsx"):
    return os.path.join(archivos_folder(dp), f"{dp}.{ext}")


def exportar_facturas(all_invoices, dp: DatePeriod, emisor_cif, rfc_prediales):
    emisor_rfc = emisor_cif['Rfc']

    emitidas = filter_invoices_iter(invoices=all_invoices.values(), fecha=dp, rfc_emisor=emisor_rfc)
    emitidas_pagos = filter_payments_iter(invoices=all_invoices, fecha=dp, rfc_emisor=emisor_rfc)
    emitidas_pagos = list(emitidas_pagos)

    recibidas = filter_invoices_iter(invoices=all_invoices.values(), fecha=dp, rfc_receptor=emisor_rfc)
    recibidas_pagos = filter_payments_iter(invoices=all_invoices, fecha=dp, rfc_receptor=emisor_rfc)

    recibidas_pagos = list(recibidas_pagos)
    pagos_hechos_iva = [
        p
        for p in recibidas_pagos
        if sum(x.get("Importe", 0) for x in p.impuestos.get("Traslados", {}).values()) > 0 and p.comprobante["Receptor"]["RegimenFiscalReceptor"] not in ('616',)
    ]
    prediales = [p for p in recibidas_pagos if p.comprobante["Emisor"]["Rfc"] in rfc_prediales]

    archivo_excel = archivos_filename(dp)
    os.makedirs(os.path.dirname(archivo_excel), exist_ok=True)

    workbook = xlsxwriter.Workbook(archivo_excel)
    # EMITIDAS
    invoices_export(workbook, "EMITIDAS", emitidas)
    payments_export(workbook, "EMITIDAS PAGOS", emitidas_pagos)

    # RECIBIDAS
    invoices_export(workbook, "RECIBIDAS", recibidas)
    payments_export(workbook, "RECIBIDAS PAGOS", recibidas_pagos)

    # SPECIALES
    payments_export(workbook, f"RECIBIDAS PAGOS IVA", pagos_hechos_iva)
    if prediales:
        payments_export(workbook, "PREDIALES", prediales)

    # RETENCIONES
    if dp.month is None:
        archivo_retenciones = archivos_filename(dp, ext="retenciones.txt")
        pagos_agrupados = payments_groupby_receptor(emitidas_pagos)
        payments_retentions_export(archivo_retenciones, pagos_agrupados)

    workbook.close()
    return archivo_excel


def calculate_declaracion_provisional(all_invoices, dp: DatePeriod, emisor_cif, rfc_prediales):
    emisor_rfc = emisor_cif['Rfc']

    emitidas_pagos = list(filter_payments_iter(invoices=all_invoices, fecha=dp, rfc_emisor=emisor_rfc))
    recibidas_pagos = list(filter_payments_iter(invoices=all_invoices, fecha=dp, rfc_receptor=emisor_rfc))
    prediales = [p for p in recibidas_pagos if p.comprobante["Emisor"]["Rfc"] in rfc_prediales]

    emitidas_pagos = sum_payments(emitidas_pagos)
    recibidas_pagos = sum_payments(recibidas_pagos)
    prediales = sum_payments(prediales)['Subtotal']

    # ISR
    total_ingresos = emitidas_pagos['Subtotal']
    deduccion_opcional = round(total_ingresos * Decimal("0.35"))
    total_deducciones = deduccion_opcional + prediales
    base_gravable = total_ingresos - deduccion_opcional - prediales
    isr_causado = isr_mensual(dp, base_gravable)
    isr_a_cargo = isr_causado - emitidas_pagos['ISR Ret']

    # IVA
    iva_a_cargo = round(total_ingresos * Decimal("0.16")) - emitidas_pagos['IVA Ret'] - recibidas_pagos['IVA16 Tras']
    res = {
        "ISR PERSONAS FÍSICAS. ARRENDAMIENTO DE INMUEBLES (USO O GOCE)": {
            "¿Tus ingresos fueron obtenidos en copropiedad o sociedad conyugal?": "No",
            "Tipo de deducción": "Deducción opcional",
            "Total de ingresos": total_ingresos,
            "Deducción opcional": deduccion_opcional,
            "Impuesto predial": prediales,
            "Total de deducciones autorizadas": total_deducciones,
            "Base gravable del pago provisional": base_gravable,
            "ISR causado": isr_causado,
            "Estímulos acreditables": 0,
            "Impuesto retenido": emitidas_pagos['ISR Ret'],
            "ISR a cargo": isr_a_cargo,
        },
        "IMPUESTO AL VALOR AGREGADO": {
            "Actividades gravadas a la tasa del 16%": total_ingresos,
            "Actividades gravadas a la tasa del 0%": 0,
            "Actividades exentas": 0,
            "Actividades no objeto del impuesto": 0,
            "IVA cobrado del periodo a la tasa del 16%": round(total_ingresos * Decimal("0.16")),
            "IVA acreditable del periodo": recibidas_pagos['IVA16 Tras'],
            "IVA retenido": emitidas_pagos['IVA Ret'],
            "¿Tienes otras cantidades a cargo?": 'No',
            "Cantidad a cargo": iva_a_cargo,
            "¿Tienes otras cantidades a favor?": 'No',
            "Impuesto a cargo": iva_a_cargo,
        },
        "TOTAL": isr_a_cargo + iva_a_cargo

    }
    p_desc = period_desc(dp)
    return p_desc + "\n" + to_yaml(res)


def sum_payments(payments):
    return {
        'Subtotal': round(sum(i.sub_total for i in payments)),
        'ISR Ret': round(sum(i.impuestos.get("Retenciones", {}).get(Impuesto.ISR, {}).get("Importe", 0) for i in payments)),
        'IVA Ret': round(sum(i.impuestos.get("Retenciones", {}).get(Impuesto.IVA, {}).get("Importe", 0) for i in payments)),

        'IVA16 Tras': round(sum(i.impuestos.get("Traslados", {}).get("002|Tasa|0.160000", {}).get("Importe", 0) for i in payments)),
    }


def isr_mensual(dp: DatePeriod, ingreso):
    table_isr_2023 = [
        (Decimal('375975.62'), Decimal("117912.32"), Decimal("0.3500")),
        (Decimal('125325.21'), Decimal("32691.18"), Decimal("0.3400")),
        (Decimal('93993.91'), Decimal("22665.17"), Decimal("0.3200")),
        (Decimal('49233.01'), Decimal("9236.89"), Decimal("0.3000")),
        (Decimal('31236.50'), Decimal("5004.12"), Decimal("0.2352")),
        (Decimal('15487.72'), Decimal("1640.18"), Decimal("0.2136")),
        (Decimal('12935.83'), Decimal("1182.88"), Decimal("0.1792")),
        (Decimal('11128.02'), Decimal("893.63"), Decimal("0.1600")),
        (Decimal('6332.06'), Decimal("371.83"), Decimal("0.1088")),
        (Decimal('746.05'), Decimal("14.32"), Decimal("0.0640")),
        (Decimal("0.00"), Decimal("0.00"), Decimal("0.0192")),
    ]

    for (limite, cuota_fija, porcentaje) in table_isr_2023:
        if ingreso >= limite:
            return round((ingreso - limite) * porcentaje + cuota_fija)


def generate_pdf_template(template_name, fields):
    template = facturacion_environment.get_template(template_name)
    md5_document = template.render(
        fields
    )
    html = markdown(md5_document)
    pdf = HTML(string=html).write_pdf(
        target=None,
        stylesheets=[
            os.path.join(SOURCE_DIRECTORY, "markdown_styles", "markdown6.css"),
            CSS(
                string='@page { width: Letter; margin: 1.6cm 1.6cm 1.6cm 1.6cm; }'
            )
        ]
    )
    return pdf


def mf_pago_fmt(cfdi):
    i = cfdi
    if i['TipoDeComprobante'] == "I":
        return i['TipoDeComprobante'].code + ' ' + i['MetodoPago'].code + ' ' + (i['FormaPago'].code if i['FormaPago'].code != '99' else '  ')
    return i['TipoDeComprobante'].code + '       '


def preview_cfdis(cfdis):
    outfile = os.path.join(TEMP_DIRECTORY, "factura.html")
    render.html_write(
        xlm=cfdis,
        target=outfile,
    )
    open_file(
        os.path.abspath(outfile)
    )


def center_location(element) -> tuple[int, int]:
    return tuple((c + s // 3) for c, s in zip(element.current_location(), element.size))



from babel.dates import format_date
from num2words import num2words

LANG = 'es_CO'


def pesos(amount):
    decimal_part = "{:.2f}".format(amount).split(".")[1]
    integer_part = num2words(int(amount), lang=LANG, to='currency').upper()

    return "${0:,.2f} (SON: {1} {2}/100M.N.)".format(amount, integer_part, decimal_part)


def num_letras(number):
    return num2words(number, lang=LANG).upper()


def numero(k):
    return str(k) + ' (' + num_letras(k) + ')'


def porcentaje(k):
    return str(k) + '% (' + num_letras(k) + ' POR CIENTO)'


def fecha(date):
    return format_date(date, locale='es_MX', format="d 'de' MMMM 'del' y").upper()


def fecha_mes(date):
    return format_date(date, locale='es_MX', format="'Mes de' MMMM 'del' y").upper()



from datetime import date, datetime, timedelta
from enum import StrEnum

import PySimpleGUI as sg
from satcfdi.models import Code
from satcfdi.catalogs import select_all
from satcfdi.pacs.sat import TipoDescargaMasivaTerceros

from satdigitalinvoice.gui_functions import mf_pago_fmt, CALENDAR_FECHA_FMT

FORMA_PAGO = select_all('C756_c_FormaPago')
TEXT_PADDING = ((5, 0), 3)

# 24 x 24
FOLDER_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAAAqElEQVR4nO3UsQnCYBRF4U/EAcRJtHUJcQddwQlsxcpVLF0hOoY2FlaC+SUQQWI0PmwscuF0j3PhFZc2" \
              "/5QR9siRnjhhid6vBVlFXGWHwS8Ft4aCHLPydozFGybo1BV8kl8wLe/muDbcryIFBwzRxbpB/OAcfV8f2y/lD17y7qcbHIPy2oKoILUFqX1RdCpSgMIVHrsUoHDVznVWM9cpQF46imlp8ye5AyE7C1To4" \
              "/HLAAAAAElFTkSuQmCC"

EXCEL_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAAA7klEQVR4nN2VQQqCUBCGv2pfFyhv0FmC7hGFq1bhrgi6gi1cBK6CdhFtvYdCi6CgrZsXwgTyUNOnLmrgBx31" \
             "/+DNOAP/HkNgCqyBK3AHLsDA1HAELIEjcANUjuYm5hbwKDBVKTnyjZ5/ATugkwWwS5qrAsBHkyyA0yBg1SQgAraiSHueCwgBL3V" \
             "/kFwWwAd6Ir8swJXcSZTEvsIRfQXEwFg6y5LruElAogDoioKma1AF4JvUoPUjctsucqi1qVfQpsY1KPuj+SY1qDsqnLaH3SJvF9Qd1wp4ilethTOT9zea7CLzsivzDPSrmPx2vAFRfA9plcmcVwAAAABJRU5ErkJggg=="

HTML_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAAA/ElEQVR4nN2VOwoCMRCGPz2HB7ARHyfwBN5EsLOxWKzs7Ox0e69gYbOtoK2dZ7BVRwZGWJbsJvuw0IGfDUkmX" \
            "/gzm8A/xgi4AC9AHNLo1gGccxaWFCAC1lUBz0CAVIVICYAYpPVNgJSF+ABtYOHoXzcF6AMdYAmsMuo1ATgZpJTvZQAScEa5ETcAiIsA+vcmDm9XgUpsjUJ7tASnGV1tTL8z26VeKTdgAOwz5esFqA6mKDX2WUR3ebe" \
            "+MXCsAkhMLsAGmNQBZC3C7MHRpxY9rO0FvMyWeUUdfIe8a6BMt9S8rqVAmlv7wZECaa43RjYx78kUh3Su5gxDAL8Vb7wE7yjidnvCAAAAAElFTkSuQmCC"

PDF_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAAA+0lEQVR4nN3VPUrEQBQA4E89gCBio7WI" \
           "+INbeAbxAoI3ESzEysoFd21tbe28QixE7a2t7dxGRwKvWCTrTpJNoQ9ek5k3X8hM5vEfo4cXfCFVZBkbbYDnCQunMeAM" \
           "/abAZyaQmiKpBpACmesSSHWRacA8Tiue92cF7GIN57j4kduzAB4DqfXdc4ACQ3xk7FEjYBk7GHQFrOMQl10BJ3GC7roC3mL8vivgNsYXsYmlyKIt8IojLOAYV1jFQWz6sA1QxOnZwv" \
           "7YFf6OG1xj1AYYRH9YwVPGT5jqXtejeMuHzMXL2tYNJ/2SZe3U6MXESS0zVWQ5t6zZywH+VnwDCQv1frFQIlYAAAAASUVORK5CYII="

ABOUT_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAABP0lEQVR4nOWVTU7DMBCFvxuwKmrX0G4o5QLtQboDFXEbwrLkGNDucwB" \
             "+VvQObS5ApCxAIz1LCNmOG68QTxopst882y/jMfwHnAIroAQ2ilJjgxzhkYRa4AvYA6" \
             "+KvcZacYx7FBZADXwC98DUw7GxAmjEXRwjbkkfwHkCfyyu5cy7yCPtxhJOUnckruUcgGGMWMqWs8D8UhE6iZ3iMVYtrTwPoVKEUEjDW123qowL" \
             "+uNSGjche6z8YlhGLHI4hGzaAi8dyVWHReiePPsmNprMXeAdeOprUZWwQB2yaKUfNM1Y4Eoa17EyLTIWeIiVqbOp0aXx4U7hw0S565RWsevRKnYpreJ3sxsniE8kntTsHObaTaN" \
             "/MvNwZvK8ETdZ3GGocnMPjom8KeofD846xZYYBuotJmQ31MK+bSzryfwb+AbymF7gpXVM1QAAAABJRU5ErkJggg=="

REFRESH_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAABGUlEQVR4nO3VzSqFURTG8Z8JYmRiJN9yDQYiKbdhYqDO9OQWyMSQDGRISoYMlHIJJEWZ" \
               "+v4oysTRrqU4nePdZej91+ptt3frWWu31/NS8kdaMINNXOANLzjFGia/ne3CdHyzGMUxarjBNpawgt0QSnsHGA6xWp1oU8bxhGvMorXBmQ5UQugO1VyBITziBD0ZxQziCh+5Aoe4z0y+ELEeyQsF0r2" \
               "/Y14etQZR2EE32jIFSv4Z7fFAfuVr3Osjh0o88ZFcgS0sRhTRF9OffEmOQBr7S/RnJO/FWUz" \
               "/QK5ANQzsOVrvbHA2DeMcbvGAsYxiflhuMr39WCfH3AurXsZOFJD2joruXcFPYwKr4a6po1ecYwNTuYlLNOMTqXpPNQBv/ywAAAAASUVORK5CYII="

PREVIEW_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAABKElEQVR4nO2VMU8CQRCFP6WAQksqOxJ/A6WJUFFprJSEilARWgsa" \
               "/4IVLYUNFZ0WVECwNxpCIDQYfwEVGj0zycNcCOTulqMxvGQze2/eztub3O3Cf0QJ6IQYD8CRi8Ed4AUUn0rTczFZGoTRfLmYRDG4Aj6BAXC8C4M80NC8H/ZNohh4K6Mel0FGu" \
               "/cPT2tjMViHvUEg9i0KxBmQjeszPVSxCnCrmBW/lUECqAKzDX/yTPmEi0EaeFauq0JDzU+BG809HX7pKAYnwAiYA9fi7qVr+jZgKEo30rqNBh2dlAe65Yx7BJLStX2teQFa4i3/JN7WBV6Z5+JrwA/wKu4S" \
               "+FYhixdADniTzvSRUQDeVXSsNiwUJ+ItbzpnpICy2vAhA4v2bLzl//ALOtSCJzC7jH4AAAAASUVORK5CYII="

SEARCH_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAABM0lEQVR4nOXTuytAYRzG8Y8RJTPJ5jJIitwGoxiZGSTZDAaDTQaJhYWBf0A" \
              "WRDHjD7BIMrkUg2wol079lEGc1yXJU2c4b8/5Pu/vfZ/Df1MjJrGKLSyjHyVfBZdjE0+4wwH2cB5rVxj8LLwmQNcYeWO3TdiOoJlUeDGOcIbqd3wFmI2QgZSAcTyiPYc3C9nBZcq" \
              "dnGAjYUPNMUVfHnNlmIcSArIpLrCUx9wSAZ3StB8V/lD1EdCTGJBVeCVvg+4wlQAvxT0m8n6wERUtyukfjakb8ga04QELOby1uMGaxFbsxq4W49jeUke05xRlKfD5gO/GJBlkGr3" \
              "ownA0JvsZD1H1Gfh8vLdiHbex/vIcYwyFX4G/VnbhdRFYIVEZbO7VmRekAn4Nnqn7nWP5NnX9JNyf1DM+Yku7BVpvYgAAAABJRU5ErkJggg=="

CONFIG_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAABPElEQVR4nOWVvWoCQRSFP5NCSJHOdGkECaa3s" \
              "/MtUu6CYOcTaGMZU2efwNYtrKwkWPgK9nmAEC38IRMGTmCZrIMru6B44MLOmXPP7N6ZvQPXgiowdcJyueEFMMBMYcTlhjfgG7hR2OfhqWYB0AEqQBnoA1/APKGZi" \
              "+tLU1GOzfXiEVirBDvgE9gA70A9oauL20izU85aHgcxAlZACxgAH0DDo29IM1DOSh6paAI/QJQyVwJCIFaE4lxE8rBe//AALIA90HbMY5VgqTDikou0lbuQVyrugIkMnsSFGttN/ENHnJ1DWqNc6+FFV" \
              "+J7jWO9tYul5pDWKJc8FxhnXSBriQKNa8eUqPBNbh5xTMeK4JRjWviP5msVkdMqnsVts7aKtGbX8zS7XtZmdwhDff5tol2/ckkXTrXoK/N88Qvnr38CSEQRlwAAAABJRU5ErkJggg=="

EDIT_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAAA10lEQVR4nO3UPUpDQRSG4cfCQgn" \
            "+LSPLSCcEC1u3EXdgEXdg3EC2IXYS0N4itZWWGkGvDHyBkE45Npq3usyB95tz5s6w4ZfZwwQveMIFtqvkB5jhDZe4xmdCyuQdxivrk3RSIl/gAe84Ta118oytCvkQPdwmZJxxXVXJl/TSSZd6O/hvs4+7yE" \
            "/Wasd4xT2ONvIl/3gsuys3dLRWGyZ0ll/2RwwiX2SnZ5XyxnkerD5uEjKqkjemmOe7H3FXJW885rGaR9xVyg/xkZBpxjXIwZewU7VTf4ovf6VSMafchm4AAAAASUVORK5CYII="

ZIP_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAAAt0lEQVR4nOXVPQrCQBCG4Qd7wTPYWtulsPAa5hq2KS0svYRnSKuNvZ3aWXgDESQibBECgWQl4s8HA8vAzMs3u" \
           "+zwi0qRR0baBJChiGhehNrGgLYq/gcwxSICsAi1n6NJsF1U7FfPz9hj1BYwDLa3ocmsBrDBBesYF32csUOvBnDCFasYwBJ3jEu5KuAYmg9iRnTDAXMkNYBGT/OVS858mpIwkraal8b4439R9i5A3tXC6Xxl" \
           "+io9AMiJawPDOx2QAAAAAElFTkSuQmCC"

IMPORT_CSV = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAABVElEQVR4nN3Vv0scYRDG8U+EQApjl4AIAYMIFqIEhGAnpLBJIRZ2goV" \
             "/gZIiIqY8rkklIb2VlWjAYq1yXGVvlzSSIlqKkZjIhoM5OJaL9+6qCTgwsDs7z3zfeff9wX20RWQJvoX+KoB3yHsU/xI5n6tA2oCUnF9VIGUA87hEE4/vAvAKH+K5kdpJGUBe8LXbAjyP0Xd6HtrkZVrWstD+f+vDW" \
             "+xiLmLj2MZHPCl0WbrjhZjPb7EEh3CEU/zAOpYiZxKf8LUMYCPEI6jFDz3DYfhYdPEb7/ET9TKAbqtoGt9xheWINWInt3KnygDehKjWIR7GaxzjIPJW43treh5EbBY7eHQdoFXsJMRNPMRmvJ" \
             "/jZeSNRqzeUfwiYvu9IAN4EcXFCCcwWMibwdNC8bwXpL1LU+0Z9grHePtYX6l6VPzNkrQpF07WxZMBqVdmVhVwE/v3gD+of4e/h7E8SwAAAABJRU5ErkJggg=="

BUTTON_COLOR = (sg.theme_background_color(), sg.theme_background_color())
LARGE_FONT = ("Courier New", 11, "bold")
PERIODO_FMT = "%Y-%m"


class TipoRecuperar(StrEnum):
    Emitidas = 'Emitidas'
    Recibidas = 'Recibidas'


class SearchOptions(StrEnum):
    PorPagar = 'Por Pagar'
    PorEnviar = 'Por Enviar'
    Hoy = datetime.now().strftime(CALENDAR_FECHA_FMT)
    Mes = datetime.now().strftime(PERIODO_FMT)
    Anio = datetime.now().strftime("%Y")


class MyTable(sg.Table):
    def __init__(self, key, headings, row_fn):
        super().__init__(
            values=[],
            key=key,
            headings=headings,
            expand_x=True,
            expand_y=True,
            select_mode=sg.TABLE_SELECT_MODE_EXTENDED,
            enable_events=True,
            text_color="black",
            background_color="white",
            # headings=HEADINGS,
            # values=[],
            # auto_size_columns=False,
            # col_widths=COL_WIDTHS,
            # justification="center",
            # num_rows=20,
            alternating_row_color="grey95",  # "aliceblue",
            # row_height=ROW_HEIGHT,
            # header_text_color="white",
            # header_background_color="darkblue",
            # font=FONT,
            # bind_return_key=True,
            # tooltip="Doble click para ver factura",
            # right_click_menu=RIGHT_CLICK_MENU
            metadata=True
        )
        self.row_fn = row_fn

    def selected_items(self):
        return [self.metadata[i] for i in self.SelectedRows]

    def select_all(self):
        self.update(
            select_rows=list(range(len(self.metadata)))
        )

    def delete_selected(self):

        def iter_metadata():
            s = 0
            for i in self.SelectedRows + [len(self.metadata)]:
                for j in range(s, i):
                    yield self.metadata[j]
                s = i + 1

        self.update(
            values=list(iter_metadata())
        )

    def update(self, values=None, **kwargs):
        if values is not None:
            super().update(
                values=[
                    self.row_fn(i, item)
                    for i, item in enumerate(values, start=1)
                ],
                **kwargs
            )
            self.metadata = values
        else:
            super().update(**kwargs)

    def refresh(self):
        self.update(
            self.metadata,
            select_rows=self.SelectedRows
        )


def make_layout(has_fiel, local_db):
    # ----- Full layout -----
    return [
        [
            sg.TabGroup(
                [[
                    sg.Tab(
                        'Emitidas '.center(13),
                        [
                            [
                                sg.Column([[
                                    sg.Button(image_data=IMPORT_CSV, key="importar_emitidas", border_width=0, button_color=BUTTON_COLOR),
                                    sg.ButtonMenu(
                                        image_data=SEARCH_ICON, button_text="", key="buscar_facturas", border_width=0, button_color=BUTTON_COLOR,
                                        menu_def=[
                                            [],
                                            [str(o) for o in SearchOptions],
                                        ],
                                    ),
                                    sg.Input(datetime.now().strftime(PERIODO_FMT), size=(40, 1), key="emitidas_search"),
                                ]],
                                    expand_x=True
                                )
                            ],
                            [
                                sg.HorizontalSeparator(color="black"),
                            ],
                            [
                                sg.Column([[
                                    sg.Button("".ljust(10), key="status_sat", border_width=0, button_color=sg.theme_background_color()),
                                    sg.Button("".ljust(10), key="email_notificada", border_width=0, button_color=sg.theme_background_color()),
                                    sg.Button("".ljust(10), key="pendiente_pago", border_width=0, button_color=sg.theme_background_color()),
                                ]]),
                                sg.VSeparator(color="black"),
                                sg.Column([[
                                    sg.CalendarButton("FechaPago:", format=CALENDAR_FECHA_FMT, title="FechaPago", no_titlebar=False, target="fecha_pago", pad=TEXT_PADDING,
                                                      border_width=0),
                                    sg.Input(datetime.now().strftime(CALENDAR_FECHA_FMT), size=(12, 1), key="fecha_pago"),
                                    sg.Text("FormaPago:", pad=TEXT_PADDING, border_width=0),
                                    sg.Combo([Code(k, v) for k, v in FORMA_PAGO.items()],
                                             default_value=Code("03", FORMA_PAGO["03"]), key="forma_pago", size=(35, 1)),
                                    sg.Text("ImpPagado:", pad=TEXT_PADDING, border_width=0),
                                    sg.Input("", size=(12, 1), key="importe_pago"),
                                ]], visible=False, key="ppd_action_items"),
                            ],
                            [
                                MyTable(
                                    key="emitidas_table",
                                    headings=[
                                        '#',
                                        'Receptor Razon Social',
                                        'Recep. Rfc',
                                        'Factura',
                                        "Fecha",
                                        "Total",
                                        "Pendiente",
                                        "Status",
                                        "Tipo",
                                        "Folio"
                                    ],
                                    row_fn=lambda i, r: [
                                        i,
                                        r['Receptor'].get('Nombre', ''),
                                        r['Receptor']['Rfc'],
                                        r.name,
                                        r["Fecha"].strftime(CALENDAR_FECHA_FMT),
                                        r["Total"],
                                        r.saldo_pendiente if r.saldo_pendiente else "",
                                        str(local_db.liquidated_state(r)) + str(" 📧" if local_db.notified(r) else "   "),
                                        mf_pago_fmt(r),
                                        r.uuid
                                    ]
                                )
                            ]],
                        key='emitidas_tab',
                    ),
                    sg.Tab(
                        'Recibidas '.center(13),
                        [
                            [
                                sg.Column([[
                                    sg.Button(image_data=IMPORT_CSV, key="importar_recibidas", border_width=0, button_color=BUTTON_COLOR),
                                    sg.ButtonMenu(
                                        image_data=SEARCH_ICON, button_text="", key="buscar_facturas_recibidas", border_width=0, button_color=BUTTON_COLOR,
                                        menu_def=[
                                            [],
                                            [str(o) for o in SearchOptions],
                                        ],
                                    ),
                                    sg.Input(datetime.now().strftime(PERIODO_FMT), size=(40, 1), key="recibidas_search"),
                                ]],
                                    expand_x=True
                                )
                            ],
                            [
                                sg.HorizontalSeparator(color="black"),
                            ],
                            [
                                sg.Column([[
                                    sg.Button("".ljust(10), key="status_sat_recibidas", border_width=0, button_color=sg.theme_background_color()),
                                ]]),
                            ],
                            [
                                MyTable(
                                    key="recibidas_table",
                                    headings=[
                                        '#',
                                        'Emisor Razon Social',
                                        'Emisor Rfc',
                                        'Factura',
                                        "Fecha",
                                        "Total",
                                        "Pendiente",
                                        "Status",
                                        "Tipo",
                                        "Folio"
                                    ],
                                    row_fn=lambda i, r: [
                                        i,
                                        r['Emisor'].get('Nombre', ''),
                                        r['Emisor']['Rfc'],
                                        r.name,
                                        r["Fecha"].strftime(CALENDAR_FECHA_FMT),
                                        r["Total"],
                                        r.saldo_pendiente if r.saldo_pendiente else "",
                                        str(local_db.liquidated_state(r)) + str(" 📧" if local_db.notified(r) else "   "),
                                        mf_pago_fmt(r),
                                        r.uuid
                                    ]
                                )
                            ]],
                        key='recibidas_tab',
                    ),
                    sg.Tab(
                        'Facturas '.center(13),
                        [
                            [
                                sg.Button(image_data=EDIT_ICON, key="editar_facturas", border_width=0, button_color=BUTTON_COLOR),
                                sg.Text("Periodo:", pad=TEXT_PADDING),
                                sg.Input(date.today().strftime(PERIODO_FMT), size=(11, 1), key="facturas_periodo"),
                                sg.Text("", pad=TEXT_PADDING, key="preparar_facturas_text", font=LARGE_FONT),
                            ],
                            [
                                MyTable(
                                    key="facturas_table",
                                    headings=[
                                        '#',
                                        'EReg',
                                        'Receptor Razon Social',
                                        'Recep. Rfc',
                                        "Tipo",
                                        "Subtotal",
                                        "Total"
                                    ],
                                    row_fn=lambda i, r: [
                                        i,
                                        r['Emisor']['RegimenFiscal'].code,
                                        r['Receptor']['Nombre'],
                                        r['Receptor']['Rfc'],
                                        mf_pago_fmt(r),
                                        r['SubTotal'],
                                        r['Total']
                                    ]
                                )
                            ]],
                        key='facturas_tab',
                    ),
                    sg.Tab(
                        'Correos'.center(13),
                        [
                            [
                                sg.Text("", pad=TEXT_PADDING),
                            ],
                            [
                                MyTable(
                                    key="correos_table",
                                    headings=[
                                        '#',
                                        'Receptor Razon Social',
                                        'Recep. Rfc',
                                        'Facturas',
                                        'Pendientes Emitidas Meses Anteriores'
                                    ],
                                    row_fn=lambda i, r: [
                                        i,
                                        r[0]["RazonSocial"],
                                        r[0]["Rfc"],
                                        ",".join(n.name for n in r[1]),
                                        ",".join(n.name for n in r[2])
                                    ]
                                )
                            ]],
                        key='correos_tab',
                    ),
                    sg.Tab(
                        'Ajustes'.center(13),
                        [
                            [
                                sg.Button(image_data=EDIT_ICON, key="editar_ajustes", border_width=0, button_color=BUTTON_COLOR),
                                sg.Text("Periodo:", pad=TEXT_PADDING),
                                sg.Input(date.today().strftime(PERIODO_FMT), size=(11, 1), key="ajustes_periodo"),
                                sg.Text("", pad=TEXT_PADDING, key="preparar_ajustes_text", font=LARGE_FONT),
                            ],
                            [
                                MyTable(
                                    key="ajustes_table",
                                    headings=[
                                        "#",
                                        "Receptor Razon Social",
                                        "Recep. Rfc",
                                        "Actual",
                                        "Nuevo",
                                        "Ajuste %",
                                        "Periodo",
                                        "Meses",
                                        "Ajuste Efectivo"
                                    ],
                                    row_fn=lambda i, r: [
                                        i,
                                        r["receptor"]["RazonSocial"],
                                        r["receptor"]["Rfc"],
                                        r["valor_unitario"],
                                        r["valor_unitario_nuevo"],
                                        r["ajuste_porcentaje"],
                                        r["periodo"],
                                        r["meses"],
                                        r["efectivo_periodo_desc"]
                                    ]
                                )
                            ]],
                        key='ajustes_tab'
                    ),
                    sg.Tab(
                        'Depositos'.center(13),
                        [
                            [
                                sg.Button(image_data=EDIT_ICON, key="editar_depositos", border_width=0, button_color=BUTTON_COLOR),
                                sg.Text("Periodo:", pad=TEXT_PADDING),
                                sg.Input(date.today().strftime(PERIODO_FMT), size=(11, 1), key="depositos_periodo"),
                                sg.Text("", pad=TEXT_PADDING, key="preparar_depositos_text", font=LARGE_FONT),
                            ],
                            [
                                MyTable(
                                    key="depositos_table",
                                    headings=[
                                        "#",
                                        "Receptor Razon Social",
                                        "Recep. Rfc",
                                        "Actual",
                                    ],
                                    row_fn=lambda i, r: [
                                        i,
                                        r["receptor"]["RazonSocial"],
                                        r["receptor"]["Rfc"],
                                        r["valor_unitario"],
                                    ]
                                )
                            ]],
                        key='depositos_tab'
                    ),
                    sg.Tab(
                        'Clientes '.center(13),
                        [
                            [
                                sg.Button(image_data=EDIT_ICON, key="editar_clientes", border_width=0, button_color=BUTTON_COLOR),
                                sg.Push(),
                                sg.Button("Exportar", key="exportar_clientes", border_width=0),
                            ],
                            [
                                MyTable(
                                    key="clientes_table",
                                    headings=[
                                        "#",
                                        "Razon Social",
                                        "Rfc",
                                        "Reg",
                                        "CP",
                                        "IdCIF"
                                    ],
                                    row_fn=lambda i, r: [
                                        i,
                                        r["RazonSocial"],
                                        r["Rfc"],
                                        r["RegimenFiscal"],
                                        r["CodigoPostal"],
                                        r["IdCIF"]
                                    ]
                                )
                            ]],
                        key='clientes_tab',
                    ),
                    sg.Tab(
                        'Solicitudes'.center(13),
                        [
                            [
                                sg.Column([[
                                    sg.Button(image_data=ZIP_ICON, key="cargar_zip", border_width=0, button_color=BUTTON_COLOR),
                                    sg.Text("Recuperar:", pad=TEXT_PADDING),
                                    sg.Combo([TipoRecuperar.Recibidas, TipoRecuperar.Emitidas], default_value=TipoRecuperar.Recibidas, key="tipo_recuperar", size=(10, 1)),

                                    sg.CalendarButton("Inicio:", format=CALENDAR_FECHA_FMT, title="Inicio", no_titlebar=False, target="fecha_inicial", pad=TEXT_PADDING,
                                                      border_width=0),
                                    sg.Input((datetime.now() - timedelta(days=40)).strftime(CALENDAR_FECHA_FMT), size=(12, 1), key="fecha_inicial"),

                                    sg.CalendarButton("Final:", format=CALENDAR_FECHA_FMT, title="Final", no_titlebar=False, target="fecha_final", pad=TEXT_PADDING,
                                                      border_width=0),
                                    sg.Input(datetime.now().strftime(CALENDAR_FECHA_FMT), size=(12, 1), key="fecha_final"),

                                    sg.Text("Tipo:", pad=TEXT_PADDING),
                                    sg.Combo([TipoDescargaMasivaTerceros.CFDI, TipoDescargaMasivaTerceros.METADATA], default_value=TipoDescargaMasivaTerceros.CFDI,
                                             key="tipo_solicitud", size=(10, 1)),

                                    sg.Button("Nueva Solicitud", key="nueva_solicitud", border_width=0),
                                ]],
                                    expand_x=True
                                )
                            ],
                            [
                                MyTable(
                                    key="solicitudes_table",
                                    headings=[
                                        "#",
                                        "IdSolicitud",
                                        "Mensaje",
                                        "EstadoSolicitud",
                                        "FechaInicial",
                                        "FechaFinal",
                                        "TipoSolicitud",
                                        "RfcReceptor",
                                        "RfcEmisor",
                                    ],
                                    row_fn=lambda i, r: [
                                        i,
                                        r["response"]["IdSolicitud"],
                                        r["response"]["Mensaje"],
                                        r["response"].get("EstadoSolicitud", ""),
                                        r["request"]["fecha_inicial"].strftime(CALENDAR_FECHA_FMT),
                                        r["request"]["fecha_final"].strftime(CALENDAR_FECHA_FMT),
                                        r["request"]["tipo_solicitud"],
                                        r["request"]["rfc_receptor"] or "",
                                        r["request"]["rfc_emisor"] or "",
                                    ]
                                )
                            ]
                        ],
                        key='solicitudes_tab',
                        # visible=has_fiel
                    ),
                    sg.Tab(
                        'Contabilidad '.center(13),
                        [
                            [
                                sg.Column([[
                                    sg.Text("Periodo:", pad=TEXT_PADDING),
                                    sg.Input(date.today().strftime(PERIODO_FMT), size=(11, 1), key="periodo"),
                                    sg.Button(image_data=EXCEL_ICON, key="ver_excel", border_width=0, button_color=BUTTON_COLOR),
                                    sg.Button(image_data=FOLDER_ICON, key="ver_carpeta", border_width=0, button_color=BUTTON_COLOR),
                                ]])
                            ],
                            [
                                sg.Multiline(
                                    expand_x=True,
                                    expand_y=True,
                                    key="declaracion",
                                    write_only=True,
                                    autoscroll=True,
                                    reroute_stdout=True
                                )
                            ]
                        ],
                        key='contabilidad_tab',
                    ),
                    sg.Tab(
                        'Consola'.center(13),
                        [
                            [
                                sg.Push(),
                                sg.Button(image_data=ABOUT_ICON, key="about", border_width=0, button_color=BUTTON_COLOR),
                            ],
                            [sg.Multiline(
                                expand_x=True,
                                expand_y=True,
                                key="console",
                                write_only=True,
                                autoscroll=True,
                                reroute_stdout=True
                            )]
                        ],
                        key='errores_tab',
                    ),
                    sg.Tab(
                        'Configuracion'.center(13),
                        [
                            [
                                sg.Column([[
                                    sg.Button(image_data=EDIT_ICON, key="editar_configurar", border_width=0, button_color=BUTTON_COLOR),
                                    sg.Button(image_data=CONFIG_ICON, key="ver_config", border_width=0, button_color=BUTTON_COLOR),
                                ]])
                            ],
                            [
                                sg.Column([[
                                    sg.Text(" Proxima Factura:", pad=TEXT_PADDING),
                                    sg.Text("Serie:", pad=TEXT_PADDING),
                                    sg.Input("", key="serie", size=(8, 1)),
                                    sg.Text("Folio:", pad=TEXT_PADDING),
                                    sg.Input("", key="folio", size=(8, 1)),
                                ]],
                                    expand_x=True
                                )
                            ],
                            [
                                sg.Column([[
                                    sg.Text("Complemento Pago:", pad=TEXT_PADDING),
                                    sg.Text("Serie:", pad=TEXT_PADDING),
                                    sg.Input("", key="serie_pago", size=(8, 1)),
                                    sg.Text("", key="folio_pago", pad=TEXT_PADDING),
                                ]],
                                    expand_x=True
                                )
                            ],
                            [
                                sg.Column([[]], size=(40, 40))
                            ],
                            [
                                sg.Column([[
                                    sg.Button("Organizar Facturas", key="organizar_facturas", border_width=0),
                                ]],
                                    expand_x=True
                                )
                            ],
                            [
                                sg.Column([[
                                    sg.Button("Exportar Metadata ", key="exportar_metadata", border_width=0),
                                    sg.Button("Importar Metadata ", key="importar_metadata", border_width=0),
                                ]],
                                    expand_x=True
                                )
                            ]
                        ],
                        key='configuracion_tab',
                    ),
                ]],
                expand_x=True,
                expand_y=True,
                enable_events=True,
                key="main_tab_group",
            )
        ],
        [
            sg.Column([[
                sg.Push(),
                sg.Text("Factura:", pad=TEXT_PADDING),
                sg.Text("", key="serie_folio", pad=TEXT_PADDING),
                sg.Button(image_data=PREVIEW_ICON, key="ver_preview", border_width=0, button_color=BUTTON_COLOR, disabled=True),
                sg.Button("".center(22), disabled=True, key="crear_facturas", border_width=0, button_color=sg.theme_background_color()),
                sg.Sizegrip(),
            ]],
                expand_x=True,
            ),
        ]
    ]


class ActionButtonManager:
    def __init__(self, button, preview):
        self.name = ""
        self.items = []
        self.button = button
        self.preview = preview

    def set_items(self, name, items):
        self.name = name
        self.items = items
        self.style_button()

    def clear(self):
        self.name = ""
        self.items = []
        self.style_button()

    def text(self):
        header = "Procesar"
        if self.name == "clientes":
            header = "Validar"
        elif self.name in ("facturas", "pago"):
            header = "Crear"
        elif self.name == "correos":
            header = "Enviar"
        elif self.name == "ajustes":
            header = "Enviar"

        if self.items:
            return f"{header} {len(self.items)} {self.name.capitalize()}"
        else:
            return ""

    def style_button(self):
        self.preview.update(
            disabled=not (bool(self.items) and self.name in ("facturas", "pago")),
        )
        self.button.update(
            self.text().center(22),
            disabled=not self.items,
            button_color=sg.theme_button_color() if self.items else sg.theme_background_color()
        )



from datetime import date
from html import escape as html_escape

import jinja2
from jinja2 import Environment, Undefined
from jinja2.filters import do_mark_safe
from satcfdi import render as cfdi_render
from satcfdi.transform.helpers import iterate as h_iterate
import satdigitalinvoice.formatting_functions.common as common

from . import TEMPLATES_DIRECTORY


class FacturacionEnvironment(Environment):
    @property
    def filter(self):
        def sub(f):
            self.filters[f.__name__] = f
            return f

        return sub

    @property
    def glob(self):
        def sub(f):
            self.globals[f.__name__] = f
            return f

        return sub

    def __init__(self):
        super().__init__(
            loader=jinja2.FileSystemLoader(searchpath=[TEMPLATES_DIRECTORY]),
            autoescape=True,
            trim_blocks=True,
            lstrip_blocks=True,
            undefined=jinja2.StrictUndefined,
        )

        @self.glob
        def iterate(v):
            if isinstance(v, Undefined):
                return v
            return h_iterate(v)

        @self.glob
        def today():
            return date.today()

        @self.filter
        def bold(k):
            return do_mark_safe(
                tag(html_escape(str(k)), "b")
            )

        @self.filter
        def moneda_nacional(k):
            return common.pesos(k)

        @self.filter
        def numero(k):
            return common.numero(k)

        @self.filter
        def porcentaje(k):
            return common.porcentaje(k)

        @self.filter
        def fecha(k):
            return common.fecha(k)

        @self.glob
        def html_str(cdfi):
            return cfdi_render.html_str(cdfi)


def tag(text, tag):
    return '<' + tag + '>' + text + '</' + tag + '>'


facturacion_environment = FacturacionEnvironment()
